# ***************************************************************************
# *   Copyright (c) 2022 Wanderer Fan <wandererfan@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "TechDrawTools package"
__author__  = "WandererFan"
__url__     = "https://www.freecad.org"
__version__ = "00.01"
__date__    = "2022-01-11"

## @package TechDrawTools
#  \ingroup TechDraw
#  \brief TechDrawTools Package for TechDraw workbench

from .TDToolsMovers import *
from .TDToolsUtil import *
from .CommandShareView import CommandShareView
from .CommandMoveView import CommandMoveView
from .CommandAxoLengthDimension import CommandAxoLengthDimension
from .CommandPositionSectionView import CommandPositionSectionView
from .CommandVertexCreations import CommandVertexCreationGroup
from .CommandHoleShaftFit import CommandHoleShaftFit
from .CommandFillTemplateFields import CommandFillTemplateFields
from .TaskShareView import TaskShareView
from .TaskMoveView import TaskMoveView
from .TaskHoleShaftFit import TaskHoleShaftFit
from .TaskAddOffsetVertex import TaskAddOffsetVertex
from .TaskFillTemplateFields import TaskFillTemplateFields



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2013 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Arch Frame"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchFrame
#  \ingroup ARCH
#  \brief The Frame object and tools
#
#  This module provides tools to build Frame objects.
#  Frames are objects made of a profile and an object with
#  edges along which the profile gets extruded

import FreeCAD
import ArchComponent
import Draft
import DraftVecUtils

if FreeCAD.GuiUp:
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond


class _Frame(ArchComponent.Component):
    "A parametric frame object"

    def __init__(self, obj):
        ArchComponent.Component.__init__(self, obj)
        self.Type = "Frame"
        self.setProperties(obj)
        obj.IfcType = "Railing"

    def setProperties(self, obj):

        pl = obj.PropertiesList
        if not "Profile" in pl:
            obj.addProperty(
                "App::PropertyLink",
                "Profile",
                "Frame",
                QT_TRANSLATE_NOOP("App::Property", "The profile used to build this frame"),
                locked=True,
            )
        if not "Align" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "Align",
                "Frame",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Specifies if the profile must be aligned with the extrusion wires",
                ),
                locked=True,
            )
            obj.Align = True
        if not "Offset" in pl:
            obj.addProperty(
                "App::PropertyVectorDistance",
                "Offset",
                "Frame",
                QT_TRANSLATE_NOOP(
                    "App::Property", "An offset vector between the base sketch and the frame"
                ),
                locked=True,
            )
        if not "BasePoint" in pl:
            obj.addProperty(
                "App::PropertyInteger",
                "BasePoint",
                "Frame",
                QT_TRANSLATE_NOOP("App::Property", "Crossing point of the path on the profile."),
                locked=True,
            )
        if not "ProfilePlacement" in pl:
            obj.addProperty(
                "App::PropertyPlacement",
                "ProfilePlacement",
                "Frame",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "An optional additional placement to add to the profile before extruding it",
                ),
                locked=True,
            )
        if not "Rotation" in pl:
            obj.addProperty(
                "App::PropertyAngle",
                "Rotation",
                "Frame",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The rotation of the profile around its extrusion axis"
                ),
                locked=True,
            )
        if not "Edges" in pl:
            obj.addProperty(
                "App::PropertyEnumeration",
                "Edges",
                "Frame",
                QT_TRANSLATE_NOOP("App::Property", "The type of edges to consider"),
                locked=True,
            )
            obj.Edges = [
                "All edges",
                "Vertical edges",
                "Horizontal edges",
                "Bottom horizontal edges",
                "Top horizontal edges",
            ]
        if not "Fuse" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "Fuse",
                "Frame",
                QT_TRANSLATE_NOOP(
                    "App::Property", "If true, geometry is fused, otherwise a compound"
                ),
                locked=True,
            )

    def onDocumentRestored(self, obj):

        ArchComponent.Component.onDocumentRestored(self, obj)
        self.setProperties(obj)

    def loads(self, state):

        self.Type = "Frame"

    def execute(self, obj):

        if self.clone(obj):
            return
        if not self.ensureBase(obj):
            return

        if not obj.Base:
            return
        if not obj.Base.Shape:
            return
        if not obj.Base.Shape.Wires:
            return

        pl = obj.Placement
        if obj.Base.Shape.Solids:
            obj.Shape = obj.Base.Shape.copy()
            if not pl.isNull():
                obj.Placement = obj.Shape.Placement.multiply(pl)
        else:
            if not obj.Profile:
                return
            if not obj.Profile.Shape:
                return
            if obj.Profile.Shape.findPlane() is None:
                return
            if not obj.Profile.Shape.Wires:
                return
            if not obj.Profile.Shape.Faces:
                for w in obj.Profile.Shape.Wires:
                    if not w.isClosed():
                        return
            import math
            import DraftGeomUtils
            import Part

            baseprofile = obj.Profile.Shape.copy()
            if hasattr(obj, "ProfilePlacement"):
                if not obj.ProfilePlacement.isNull():
                    baseprofile.Placement = obj.ProfilePlacement.multiply(baseprofile.Placement)
            if not baseprofile.Faces:
                f = []
                for w in baseprofile.Wires:
                    f.append(Part.Face(w))
                if len(f) == 1:
                    baseprofile = f[0]
                else:
                    baseprofile = Part.makeCompound(f)
            shapes = []
            normal = DraftGeomUtils.getNormal(obj.Base.Shape)
            edges = obj.Base.Shape.Edges
            if hasattr(obj, "Edges"):
                if obj.Edges == "Vertical edges":
                    rv = obj.Base.Placement.Rotation.multVec(FreeCAD.Vector(0, 1, 0))
                    edges = [
                        e
                        for e in edges
                        if round(rv.getAngle(e.tangentAt(e.FirstParameter)), 4) in [0, 3.1416]
                    ]
                elif obj.Edges == "Horizontal edges":
                    rv = obj.Base.Placement.Rotation.multVec(FreeCAD.Vector(1, 0, 0))
                    edges = [
                        e
                        for e in edges
                        if round(rv.getAngle(e.tangentAt(e.FirstParameter)), 4) in [0, 3.1416]
                    ]
                elif obj.Edges == "Top horizontal edges":
                    rv = obj.Base.Placement.Rotation.multVec(FreeCAD.Vector(1, 0, 0))
                    edges = [
                        e
                        for e in edges
                        if round(rv.getAngle(e.tangentAt(e.FirstParameter)), 4) in [0, 3.1416]
                    ]
                    edges = sorted(edges, key=lambda x: x.CenterOfMass.z, reverse=True)
                    z = edges[0].CenterOfMass.z
                    edges = [e for e in edges if abs(e.CenterOfMass.z - z) < 0.00001]
                elif obj.Edges == "Bottom horizontal edges":
                    rv = obj.Base.Placement.Rotation.multVec(FreeCAD.Vector(1, 0, 0))
                    edges = [
                        e
                        for e in edges
                        if round(rv.getAngle(e.tangentAt(e.FirstParameter)), 4) in [0, 3.1416]
                    ]
                    edges = sorted(edges, key=lambda x: x.CenterOfMass.z)
                    z = edges[0].CenterOfMass.z
                    edges = [e for e in edges if abs(e.CenterOfMass.z - z) < 0.00001]
            for e in edges:
                bvec = DraftGeomUtils.vec(e)
                bpoint = e.Vertexes[0].Point
                profile = baseprofile.copy()
                rot = None  # New rotation.
                # Supplying FreeCAD.Rotation() with two parallel vectors and
                # a null vector may seem strange, but the function is perfectly
                # able to handle this. Its algorithm will use default axes in
                # such cases.
                if obj.Align:
                    if normal is None:
                        rot = FreeCAD.Rotation(FreeCAD.Vector(), bvec, bvec, "ZYX")
                    else:
                        rot = FreeCAD.Rotation(FreeCAD.Vector(), normal, bvec, "ZYX")
                    profile.Placement.Rotation = rot
                if hasattr(obj, "BasePoint"):
                    edges = Part.__sortEdges__(profile.Edges)
                    basepointliste = [profile.Placement.Base]
                    for edge in edges:
                        basepointliste.append(DraftGeomUtils.findMidpoint(edge))
                        basepointliste.append(edge.Vertexes[-1].Point)
                    try:
                        basepoint = basepointliste[obj.BasePoint]
                    except IndexError:
                        FreeCAD.Console.PrintMessage(
                            translate("Arch", "Crossing point not found in profile.") + "\n"
                        )
                        basepoint = basepointliste[0]
                else:
                    basepoint = profile.Placement.Base
                delta = bpoint.sub(basepoint)  # Translation vector.
                if obj.Offset and (not DraftVecUtils.isNull(obj.Offset)):
                    if rot is None:
                        delta = delta + obj.Offset
                    else:
                        delta = delta + rot.multVec(obj.Offset)
                profile.translate(delta)
                if obj.Rotation:
                    profile.rotate(bpoint, bvec, obj.Rotation)
                # profile = wire.makePipeShell([profile], True, False, 2) TODO buggy
                profile = profile.extrude(bvec)
                shapes.append(profile)
            if shapes:
                if hasattr(obj, "Fuse"):
                    if obj.Fuse:
                        if len(shapes) > 1:
                            s = shapes[0].multiFuse(shapes[1:])
                            s = s.removeSplitter()
                            obj.Shape = s
                            obj.Placement = pl
                            return
                obj.Shape = Part.makeCompound(shapes)
                obj.Placement = pl


class _ViewProviderFrame(ArchComponent.ViewProviderComponent):
    "A View Provider for the Frame object"

    def __init__(self, vobj):

        ArchComponent.ViewProviderComponent.__init__(self, vobj)

    def getIcon(self):

        import Arch_rc

        return ":/icons/Arch_Frame_Tree.svg"

    def claimChildren(self):

        p = []
        if hasattr(self, "Object"):
            if self.Object.Profile:
                p = [self.Object.Profile]
        return ArchComponent.ViewProviderComponent.claimChildren(self) + p



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM ShapeView command"""

import FreeCAD
import FreeCADGui

from draftguitools import gui_shape2dview

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate


class BIM_Shape2DView(gui_shape2dview.Shape2DView):

    def GetResources(self):
        d = super().GetResources()
        d["Pixmap"] = "Arch_BuildingPart_Tree"
        d["MenuText"] = QT_TRANSLATE_NOOP("BIM_Shape2DView", "Section View")
        d["Accel"] = "V, V"
        return d

    def proceed(self):
        """Proceed with the command if one object was selected."""
        self.proceed_BimShape2DView()  # Common
        if self.commitlist_BimShape2DView:
            commitlist = self.commitlist_BimShape2DView
            commitlist.append("FreeCAD.ActiveDocument.recompute()")
            self.commit(translate("draft", "Create 2D view"), self.commitlist_BimShape2DView)
        self.finish()

    def proceed_BimShape2DView(self):  # Common
        """Proceed with the command if one object was selected."""
        # difference from Draft: it sets InPlace to False
        import DraftVecUtils

        if self.call is not None:
            self.end_callbacks(self.call)
        faces = []
        objs = []
        view = FreeCADGui.ActiveDocument.ActiveView
        vec = view.getViewDirection().negative()
        sel = FreeCADGui.Selection.getSelectionEx()
        for s in sel:
            objs.append(s.Object)
            for e in s.SubElementNames:
                if "Face" in e:
                    faces.append(int(e[4:]) - 1)
        # print(objs, faces)
        self.objs_BimShape2DView = objs
        commitlist = []
        FreeCADGui.addModule("Draft")
        if len(objs) == 1 and faces:
            _cmd = "Draft.make_shape2dview"
            _cmd += "("
            _cmd += "FreeCAD.ActiveDocument." + objs[0].Name + ", "
            _cmd += DraftVecUtils.toString(vec) + ", "
            _cmd += "facenumbers=" + str(faces)
            _cmd += ")"
            # commitlist.append("sv = " + _cmd)
            # commitlist.append("sv.InPlace = False")
            commitlist.append("sv0 = " + _cmd)
            commitlist.append("sv0.InPlace = False")
        else:
            n = 0
            for o in objs:
                _cmd = "Draft.make_shape2dview"
                _cmd += "("
                _cmd += "FreeCAD.ActiveDocument." + o.Name + ", "
                _cmd += DraftVecUtils.toString(vec)
                _cmd += ")"
                commitlist.append("sv" + str(n) + " = " + _cmd)
                commitlist.append("sv" + str(n) + ".InPlace = False")
                n += 1
        if commitlist:
            # commitlist.append("FreeCAD.ActiveDocument.recompute()")
            self.commitlist_BimShape2DView = commitlist
        else:
            self.commitlist_BimShape2DView = None

        #    self.commit(translate("draft", "Create 2D view"),
        #                commitlist)
        # self.finish()


class BIM_Shape2DCut(BIM_Shape2DView):

    def GetResources(self):
        d = super().GetResources()
        d["Pixmap"] = "Arch_View_Cut"
        d["MenuText"] = QT_TRANSLATE_NOOP("BIM_Shape2DView", "Section Cut")
        d["Accel"] = "V, C"
        return d

    def proceed(self):
        # super().proceed()
        #'sv' is not passed from BIM_Shape2DView() to BIM_Shape2DCut()
        # FreeCADGui.doCommand("sv.ProjectionMode = \"Cutfaces\"")

        """Proceed with the command if one object was selected."""
        self.proceed_BimShape2DView()  # Common
        if self.commitlist_BimShape2DView:
            commitlist = self.commitlist_BimShape2DView

            # BIM_Shape2DCut specific
            # commitlist.append("sv.ProjectionMode = \"Cutfaces\"")
            objs = self.objs_BimShape2DView
            n = 0
            for o in objs:
                commitlist.append("sv" + str(n) + '.ProjectionMode = "Cutfaces"')
                n += 1
            commitlist.append("FreeCAD.ActiveDocument.recompute()")
            self.commit(translate("draft", "Create 2D Cut"), commitlist)
        self.finish()


FreeCADGui.addCommand("BIM_Shape2DView", BIM_Shape2DView())
FreeCADGui.addCommand("BIM_Shape2DCut", BIM_Shape2DCut())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2018 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Arch External Reference"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchReference
#  \ingroup ARCH
#  \brief The Reference object and tools
#
#  This module provides tools to build Reference objects.
#  References can take a shape from a Part-based object in
#  another file.

import os
import re
import struct
import zipfile

import FreeCAD

from draftutils import params

if FreeCAD.GuiUp:
    from PySide import QtCore, QtGui
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond


class ArchReference:
    """The Arch Reference object"""

    def __init__(self, obj):

        obj.Proxy = self
        self.Type = "Reference"
        ArchReference.setProperties(self, obj)
        self.reload = True

    def setProperties(self, obj):

        pl = obj.PropertiesList
        if not "File" in pl:
            t = QT_TRANSLATE_NOOP("App::Property", "The base file this component is built upon")
            obj.addProperty("App::PropertyFile", "File", "Reference", t, locked=True)
        if not "Part" in pl:
            t = QT_TRANSLATE_NOOP("App::Property", "The part to use from the base file")
            obj.addProperty("App::PropertyString", "Part", "Reference", t, locked=True)
        if not "ReferenceMode" in pl:
            t = QT_TRANSLATE_NOOP(
                "App::Property",
                "The way the referenced objects are included in the current document. 'Normal' includes the shape, 'Transient' discards the shape when the object is switched off (smaller filesize), 'Lightweight' does not import the shape but only the OpenInventor representation",
            )
            obj.addProperty(
                "App::PropertyEnumeration", "ReferenceMode", "Reference", t, locked=True
            )
            obj.ReferenceMode = ["Normal", "Transient", "Lightweight"]
            if "TransientReference" in pl:
                if obj.TransientReference:
                    obj.ReferenceMode = "Transient"
                obj.removeProperty("TransientReference")
                t = translate("Arch", "TransientReference property to ReferenceMode")
                FreeCAD.Console.PrintMessage(
                    translate("Arch", "Upgrading") + " " + obj.Label + " " + t + "\n"
                )
        if not "FuseArch" in pl:
            t = QT_TRANSLATE_NOOP("App::Property", "Fuse objects of same material")
            obj.addProperty("App::PropertyBool", "FuseArch", "Reference", t, locked=True)

    def onDocumentRestored(self, obj):

        ArchReference.setProperties(self, obj)
        if obj.ReferenceMode == "Lightweight":
            self.reload = False
            if obj.ViewObject and obj.ViewObject.Proxy:
                obj.ViewObject.Proxy.loadInventor(obj)
        else:
            self.reload = True
            self.execute(obj)  # sets self.reload to False again

    def dumps(self):

        return None

    def loads(self, state):

        self.Type = "Reference"

    def onChanged(self, obj, prop):

        if prop in ["File", "Part"]:
            self.reload = True
        elif prop == "ReferenceMode":
            if obj.ReferenceMode == "Normal":
                if obj.ViewObject and obj.ViewObject.Proxy:
                    obj.ViewObject.Proxy.unloadInventor(obj)
                if (not obj.Shape) or obj.Shape.isNull():
                    self.reload = True
                    obj.touch()
            elif obj.ReferenceMode == "Transient":
                if obj.ViewObject and obj.ViewObject.Proxy:
                    obj.ViewObject.Proxy.unloadInventor(obj)
                self.reload = False
            elif obj.ReferenceMode == "Lightweight":
                self.reload = False
                import Part

                pl = obj.Placement
                obj.Shape = Part.Shape()
                obj.Placement = pl
                if obj.ViewObject and obj.ViewObject.Proxy:
                    obj.ViewObject.Proxy.loadInventor(obj)

    def execute(self, obj):

        import Part

        filename = self.getFile(obj)
        if filename and self.reload and obj.ReferenceMode in ["Normal", "Transient"]:
            self.parts = self.getPartsList(obj)
            if self.parts:
                if filename.lower().endswith(".fcstd"):
                    zdoc = zipfile.ZipFile(filename)
                    if zdoc:
                        pl = obj.Placement
                        self.shapes = []
                        if obj.Part:
                            if obj.Part in self.parts:
                                if self.parts[obj.Part][1] in zdoc.namelist():
                                    f = zdoc.open(self.parts[obj.Part][1])
                                    shapedata = f.read()
                                    f.close()
                                    shapedata = shapedata.decode("utf8")
                                    shape = self.cleanShape(shapedata, obj, self.parts[obj.Part][2])
                                    self.shapes.append(shape)
                                    obj.Shape = shape
                                else:
                                    t = translate("Arch", "Part not found in file")
                                    FreeCAD.Console.PrintError(t + "\n")
                        else:
                            for part in self.parts.values():
                                if part[3]:
                                    # Do not include BuildingParts as their
                                    # shape is just a copy of their group:
                                    continue
                                f = zdoc.open(part[1])
                                shapedata = f.read()
                                f.close()
                                shapedata = shapedata.decode("utf8")
                                shape = self.cleanShape(shapedata, obj)
                                self.shapes.append(shape)
                            if self.shapes:
                                obj.Shape = Part.makeCompound(self.shapes)
                        obj.Placement = pl
                elif filename.lower().endswith(".ifc"):
                    ifcfile = self.getIfcFile(filename)
                    if not ifcfile:
                        return
                    try:
                        from nativeifc import ifc_tools
                        from nativeifc import ifc_generator
                    except:
                        t = translate(
                            "Arch", "NativeIFC not available - unable to process IFC files"
                        )
                        FreeCAD.Console.PrintError(t + "\n")
                        return
                    elements = self.getIFCElements(obj, ifcfile)
                    shape, colors = ifc_generator.generate_shape(ifcfile, elements, cached=True)
                    if shape:
                        placement = shape.Placement
                        obj.Shape = shape
                        obj.Placement = placement
                        if colors:
                            ifc_tools.set_colors(obj, colors)
            elif filename.lower().endswith(".dxf"):
                # create a special parameter set to control the DXF importer
                loc = "User parameter:BaseApp/Preferences/Mod/Arch"
                hGrp = FreeCAD.ParamGet(loc).GetGroup("RefDxfImport")
                hGrp.SetBool("dxfUseDraftVisGroups", False)
                hGrp.SetBool("dxfGetOriginalColors", False)
                hGrp.SetBool("groupLayers", True)
                hGrp.SetFloat("dxfScaling", 1.0)
                hGrp.SetBool("dxftext", False)
                hGrp.SetBool("dxfImportPoints", True)
                hGrp.SetBool("dxflayout", False)
                hGrp.SetBool("dxfstarblocks", False)
                doc = obj.Document
                oldobjs = list(doc.Objects)
                import Import

                Import.readDXF(filename, doc.Name, True, loc + "/RefDxfImport")
                newobjs = [o for o in doc.Objects if o not in oldobjs]
                shapes = [o.Shape for o in newobjs if o.isDerivedFrom("Part::Feature")]
                if len(shapes) == 1:
                    obj.Shape = shapes[0]
                elif len(shapes) > 1:
                    obj.Shape = Part.makeCompound(shapes)
                names = [o.Name for o in newobjs]
                for n in names:
                    doc.removeObject(n)
            self.reload = False

    def getIFCElements(self, obj, ifcfile):
        """returns IFC elements for this object"""

        try:
            from nativeifc import ifc_generator
        except:
            t = translate("Arch", "NativeIFC not available - unable to process IFC files")
            FreeCAD.Console.PrintError(t + "\n")
            return
        if obj.Part:
            element = ifcfile[int(obj.Part)]
        else:
            element = ifcfile.by_type("IfcProject")[0]
        elements = ifc_generator.get_decomposed_elements(element)
        elements = ifc_generator.filter_types(elements)
        return elements

    def cleanShape(self, shapedata, obj, materials=None):
        """cleans the imported shape"""

        import Part

        shape = Part.Shape()
        shape.importBrepFromString(shapedata)
        if obj.FuseArch and materials:
            # separate lone edges
            shapes = []
            for edge in shape.Edges:
                found = False
                for solid in shape.Solids:
                    for soledge in solid.Edges:
                        if edge.hashCode() == soledge.hashCode():
                            found = True
                            break
                    if found:
                        break
                if found:
                    break
            else:
                shapes.append(edge)
            # print("solids:",len(shape.Solids),"mattable:",materials)
            for key, solindexes in materials.items():
                if key == "Undefined":
                    # do not join objects with no defined material
                    for solindex in [int(i) for i in solindexes.split(",")]:
                        shapes.append(shape.Solids[solindex])
                else:
                    fusion = None
                    for solindex in [int(i) for i in solindexes.split(",")]:
                        if not fusion:
                            fusion = shape.Solids[solindex]
                        else:
                            fusion = fusion.fuse(shape.Solids[solindex])
                    if fusion:
                        shapes.append(fusion)
            shape = Part.makeCompound(shapes)
            try:
                shape = shape.removeSplitter()
            except Exception:
                t = translate("Arch", "Error removing splitter")
                FreeCAD.Console.PrintError(obj.Label + ": " + t + "\n")
        return shape

    def exists(self, filepath):
        """case-insensitive version of os.path.exists. Returns the actual file path or None"""

        if os.path.exists(filepath):
            return filepath
        # check for uppercase/lowercase extensions
        p, e = os.path.splitext(filepath)
        if os.path.exists(p + e.lower()):
            return p + e.lower()
        if os.path.exists(p + e.upper()):
            return p + e.upper()
        return None

    def getFile(self, obj, filename=None):
        """gets a valid file, if possible"""

        if not filename:
            filename = obj.File
        if not filename:
            return None
        if not filename.lower().endswith(".fcstd"):
            if not filename.lower().endswith(".ifc"):
                if not filename.lower().endswith(".dxf"):
                    return None
        if not self.exists(filename):
            # search for the file in the current directory if not found
            basename = os.path.basename(filename)
            currentdir = os.path.dirname(obj.Document.FileName)
            altfile = os.path.join(currentdir, basename)
            if altfile == obj.Document.FileName:
                return None
            elif self.exists(altfile):
                return self.exists(altfile)
            else:
                # search for subpaths in current folder
                altfile = None
                subdirs = self.splitall(os.path.dirname(filename))
                for i in range(len(subdirs)):
                    subpath = [currentdir] + subdirs[-i:] + [basename]
                    altfile = os.path.join(*subpath)
                    if self.exists(altfile):
                        return self.exists(altfile)
                return None
        return self.exists(filename)

    def getPartsList(self, obj, filename=None):
        """returns a list of Part-based objects in a file"""

        filename = self.getFile(obj, filename)
        if not filename:
            return None
        if filename.lower().endswith(".fcstd"):
            return self.getPartsListFCSTD(obj, filename)
        elif filename.lower().endswith(".ifc"):
            return self.getPartsListIFC(obj, filename)
        elif filename.lower().endswith(".dxf"):
            return self.getPartsListDXF(obj, filename)

    def getPartsListDXF(self, obj, filename):
        """returns a list of Part-based objects in a DXF file"""

        # support layers
        # with open(filename) as f:
        #    txt = f.read()
        return {}

    def getPartsListIFC(self, obj, filename):
        """returns a list of Part-based objects in a IFC file"""

        ifcfile = self.getIfcFile(filename)
        if not ifcfile:
            return None
        structs = ifcfile.by_type("IfcSpatialElement")
        res = {}
        for s in structs:
            n = s.Name
            if not n:
                n = ""
            name = "#" + str(s.id()) + " " + n + "(" + s.is_a() + ")"
            res[str(s.id())] = [name, s, None]
        return res

    def getPartsListFCSTD(self, obj, filename):
        """returns a list of Part-based objects in a FCStd file"""

        parts = {}
        materials = {}
        zdoc = zipfile.ZipFile(filename)
        with zdoc.open("Document.xml") as docf:
            name = None
            label = None
            part = None
            materials = {}
            is_buildingpart = False
            writemode = False
            for line in docf:
                line = line.decode("utf8")
                if "<Object name=" in line:
                    n = re.findall(r"name=\"(.*?)\"", line)
                    if n:
                        name = n[0]
                elif 'class="BuildingPart"' in line:
                    is_buildingpart = True
                elif '<Property name="Label"' in line:
                    writemode = True
                elif writemode and "<String value=" in line:
                    n = re.findall(r"value=\"(.*?)\"", line)
                    if n:
                        label = n[0]
                        writemode = False
                elif '<Property name="Shape" type="Part::PropertyPartShape"' in line:
                    writemode = True
                elif writemode and "<Part" in line and "file=" in line:
                    n = re.findall(r"file=\"(.*?)\"", line)
                    if n:
                        part = n[0]
                        writemode = False
                elif '<Property name="MaterialsTable" type="App::PropertyMap"' in line:
                    writemode = True
                elif writemode and "<Item key=" in line:
                    n = re.findall(r"key=\"(.*?)\"", line)
                    v = re.findall(r"value=\"(.*?)\"", line)
                    if n and v:
                        materials[n[0]] = v[0]
                elif writemode and "</Map>" in line:
                    writemode = False
                elif "</Object>" in line:
                    if name and label and part:
                        parts[name] = [label, part, materials, is_buildingpart]
                    name = None
                    label = None
                    part = None
                    materials = {}
                    is_buildingpart = False
                    writemode = False
        return parts

    def getIfcFile(self, filename):
        """Gets an IfcOpenShell object"""

        try:
            import ifcopenshell
        except:
            t = translate("Arch", "NativeIFC not available - unable to process IFC files")
            FreeCAD.Console.PrintError(t + "\n")
            return None
        if not getattr(self, "ifcfile", None):
            self.ifcfile = ifcopenshell.open(filename)
        return self.ifcfile

    def getColors(self, obj):
        """returns the Shape Appearance of the referenced object(s)"""

        filename = self.getFile(obj)
        if not filename:
            return []
        if not filename.lower().endswith(".fcstd"):
            return []
        if not getattr(self, "parts", {}):
            return []
        if not getattr(self, "shapes", []):
            return []

        totalcolors = []
        if obj.Part:
            parts = [obj.Part]
        else:
            # Do not include BuildingParts as their shape is just a copy of their group:
            parts = [key for key, val in self.parts.items() if not val[3]]
        lenparts = len(parts)
        for i, part in enumerate(parts):
            lenfaces = len(self.shapes[i].Faces)
            if lenfaces:
                colors = self._getColorsPart(filename, part)
                if len(colors) == lenfaces:
                    totalcolors.extend(colors)
                elif lenparts == 1:
                    totalcolors.append(colors[0])
                else:
                    totalcolors.extend([colors[0]] * lenfaces)

        return totalcolors

    def _getColorsPart(self, filename, part):
        zdoc = zipfile.ZipFile(filename)
        if not "GuiDocument.xml" in zdoc.namelist():
            return []
        colorfile = None
        with zdoc.open("GuiDocument.xml") as docf:
            writemode1 = False
            writemode2 = False
            writemode3 = False
            for line in docf:
                line = line.decode("utf8")
                if ('<ViewProvider name="' + part + '"') in line:
                    writemode1 = True
                elif writemode1 and '<ViewProvider name="' in line:
                    # We have reached the next item:
                    break
                elif writemode1 and ('<Property name="DiffuseColor"' in line):
                    writemode2 = True
                elif writemode1 and ('<Property name="ShapeAppearance"' in line):
                    writemode3 = True
                elif writemode2 and ("<ColorList file=" in line):
                    n = re.findall(r"file=\"(.*?)\"", line)
                    if n and n[0] and n[0] in zdoc.namelist():
                        colorfile = n[0]
                        writemode3 = False
                        break
                elif writemode3 and ("<MaterialList file=" in line):
                    n = re.findall(r"file=\"(.*?)\"", line)
                    if n and n[0] and n[0] in zdoc.namelist():
                        colorfile = n[0]
                        writemode2 = False
                        break

        if colorfile is None:
            return []

        cf = zdoc.open(colorfile)
        buf = cf.read()
        cf.close()
        colors = []

        if writemode2:
            # Old DiffuseColor support:
            alpha_to_transparency = False
            for i in range(1, int(len(buf) / 4)):
                # ShapeAppearance material with default v0.21 properties:
                material = FreeCAD.Material()
                color = self._processColor(buf, i * 4)
                material.DiffuseColor = color[:3] + (255,)
                material.Transparency = color[3] / 255.0
                colors.append(material)
                if material.Transparency == 1.0:
                    # Assumption: a face with 100% transparency indicates
                    # we are actually dealing with alpha values.
                    alpha_to_transparency = True
            if alpha_to_transparency:
                for material in colors:
                    material.Transparency = 1.0 - material.Transparency

        elif writemode3:
            # File format ShapeAppearance files in FCStd file:
            # - 1st byte: number of faces
            # - Next 3 bytes: zero
            # - For each face: 6 fields with 4 bytes in this order:
            #   - ambient color
            #   - diffuse color
            #   - specular color
            #   - emissive color
            #   - shininess
            #   - transparency
            # - Tail of file: unknown repetition of bytes
            for face_idx in range(buf[0]):
                face_buf = buf[face_idx * 24 + 4 : face_idx * 24 + 28]
                # ShapeAppearance material with default v0.21 properties:
                material = FreeCAD.Material()
                material.AmbientColor = self._processColor(face_buf, 0)
                material.DiffuseColor = self._processColor(face_buf, 4)
                material.SpecularColor = self._processColor(face_buf, 8)
                material.EmissiveColor = self._processColor(face_buf, 12)
                material.Shininess = self._processNumber(face_buf, 16)
                material.Transparency = self._processNumber(face_buf, 20)
                colors.append(material)

        return colors

    def _processColor(self, buf, i):
        """returns a tuple with 4 ints (0-255)"""
        return (buf[i + 3], buf[i + 2], buf[i + 1], buf[i])

    def _processNumber(self, buf, i):
        """returns a float"""
        return struct.unpack("f", buf[i : i + 4])[0]

    def splitall(self, path):
        """splits a path between its components"""

        allparts = []
        while 1:
            parts = os.path.split(path)
            if parts[0] == path:  # sentinel for absolute paths
                allparts.insert(0, parts[0])
                break
            elif parts[1] == path:  # sentinel for relative paths
                allparts.insert(0, parts[1])
                break
            else:
                path = parts[0]
                allparts.insert(0, parts[1])
        return allparts


class ViewProviderArchReference:
    """A View Provider for the Arch Reference object"""

    def __init__(self, vobj):

        vobj.Proxy = self
        self.setProperties(vobj)

    def setProperties(self, vobj):

        pl = vobj.PropertiesList
        if not "TimeStamp" in pl:
            t = QT_TRANSLATE_NOOP("App::Property", "The latest time stamp of the linked file")
            vobj.addProperty("App::PropertyFloat", "TimeStamp", "Reference", t, locked=True)
            vobj.setEditorMode("TimeStamp", 2)
        if not "UpdateColors" in pl:
            t = QT_TRANSLATE_NOOP(
                "App::Property", "If true, the colors from the linked file will be kept updated"
            )
            vobj.addProperty("App::PropertyBool", "UpdateColors", "Reference", t, locked=True)
            vobj.UpdateColors = True

    def getIcon(self):

        import Arch_rc

        return ":/icons/Arch_Reference.svg"

    def attach(self, vobj):

        self.Object = vobj.Object
        # Check for file change every minute
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.checkChanges)
        s = params.get_param_arch("ReferenceCheckInterval")
        self.timer.start(1000 * s)

    def dumps(self):

        return None

    def loads(self, state):

        return None

    def updateData(self, obj, prop):
        if prop == "Shape":
            if hasattr(obj.ViewObject, "UpdateColors") and obj.ViewObject.UpdateColors:
                if obj.Shape and not obj.Shape.isNull():
                    colors = obj.Proxy.getColors(obj)
                    if colors:
                        obj.ViewObject.ShapeAppearance = colors

    def checkChanges(self):
        "checks if the linked file has changed"

        if hasattr(self, "Object") and self.Object:
            try:
                f = self.Object.File
            except ReferenceError:
                f = None
                if hasattr(self, "timer"):
                    self.timer.stop()
                    del self.timer
            if f:
                filename = self.Object.Proxy.getFile(self.Object)
                if filename:
                    st_mtime = os.stat(filename).st_mtime
                    if hasattr(self.Object.ViewObject, "TimeStamp"):
                        if self.Object.ViewObject.TimeStamp:
                            if self.Object.ViewObject.TimeStamp != st_mtime:
                                self.Object.Proxy.reload = True
                                self.Object.touch()
                        self.Object.ViewObject.TimeStamp = st_mtime

    def onChanged(self, vobj, prop):

        if prop == "ShapeColor":
            # prevent ShapeColor from overriding DiffuseColor
            if hasattr(vobj, "DiffuseColor") and hasattr(vobj, "UpdateColors"):
                if vobj.DiffuseColor and vobj.UpdateColors:
                    vobj.DiffuseColor = vobj.DiffuseColor
        elif prop == "Visibility":
            if vobj.Visibility:
                if (not vobj.Object.Shape) or vobj.Object.Shape.isNull():
                    vobj.Object.Proxy.reload = True
                    vobj.Object.Proxy.execute(vobj.Object)
            else:
                if hasattr(vobj.Object, "ReferenceMode"):
                    if vobj.Object.ReferenceMode == "Transient":
                        vobj.Object.Proxy.reload = False
                        import Part

                        pl = vobj.Object.Placement
                        vobj.Object.Shape = Part.Shape()
                        vobj.Object.Placement = pl

    def onDelete(self, obj, doc):

        if hasattr(self, "timer"):
            self.timer.stop()
            del self.timer
            return True

    def setEdit(self, vobj, mode):
        if mode != 0:
            return None

        taskd = ArchReferenceTaskPanel(vobj.Object)
        FreeCADGui.Control.showDialog(taskd)
        return True

    def unsetEdit(self, vobj, mode):
        if mode != 0:
            return None

        FreeCADGui.Control.closeDialog()
        return True

    def setupContextMenu(self, vobj, menu):

        if FreeCADGui.activeWorkbench().name() != "BIMWorkbench":
            return

        actionEdit = QtGui.QAction(translate("Arch", "Edit"), menu)
        QtCore.QObject.connect(actionEdit, QtCore.SIGNAL("triggered()"), self.edit)
        menu.addAction(actionEdit)

        actionOnReload = QtGui.QAction(
            QtGui.QIcon(":/icons/view-refresh.svg"), translate("Arch", "Reload reference"), menu
        )
        QtCore.QObject.connect(actionOnReload, QtCore.SIGNAL("triggered()"), self.onReload)
        menu.addAction(actionOnReload)

        actionOnOpen = QtGui.QAction(
            QtGui.QIcon(":/icons/document-open.svg"), translate("Arch", "Open reference"), menu
        )
        QtCore.QObject.connect(actionOnOpen, QtCore.SIGNAL("triggered()"), self.onOpen)
        menu.addAction(actionOnOpen)

    def edit(self):

        FreeCADGui.ActiveDocument.setEdit(self.Object, 0)

    def onReload(self):
        "reloads the reference object"

        if hasattr(self, "Object") and self.Object:
            self.Object.Proxy.reload = True
            self.Object.touch()
            FreeCAD.ActiveDocument.recompute()

    def onOpen(self):
        "opens the reference file"

        if hasattr(self, "Object") and self.Object:
            if self.Object.File:
                FreeCAD.openDocument(self.Object.File)

    def loadInventor(self, obj):
        "loads an openinventor file and replace the root node of this object"

        filename = obj.Proxy.getFile(obj)
        if not filename:
            return None
        if filename.lower().endswith(".ifc"):
            self.setIFCNode(obj, filename)
            return

        # check inventor contents
        ivstring = self.getInventorString(obj)
        if not ivstring:
            t = translate("Arch", "Unable to get lightWeight node for object referenced in")
            FreeCAD.Console.PrintWarning(t + " " + obj.Label + "\n")
            return
        from pivy import coin
        from draftutils import gui_utils

        inputnode = coin.SoInput()
        inputnode.setBuffer(ivstring)
        lwnode = coin.SoDB.readAll(inputnode)
        if not isinstance(lwnode, coin.SoSeparator):
            t = translate("Arch", "Invalid lightWeight node for object referenced in")
            FreeCAD.Console.PrintError(t + " " + obj.Label + "\n")
            return
        if lwnode.getNumChildren() < 2:
            t = translate("Arch", "Invalid lightWeight node for object referenced in")
            FreeCAD.Console.PrintError(t + " " + obj.Label + "\n")
            return
        flatlines = lwnode
        shaded = lwnode.getChild(0)
        wireframe = lwnode.getChild(1)

        # check node contents
        rootnode = obj.ViewObject.RootNode
        # display mode switch
        switch = gui_utils.find_coin_node(obj.ViewObject.RootNode, coin.SoSwitch)
        if switch is None or switch.getNumChildren() != 4:
            FreeCAD.Console.PrintError(
                translate("Arch", "Invalid root node in") + " " + obj.Label + "\n"
            )
            return

        # keep a copy of the original nodes
        self.orig_flatlines = switch.getChild(0).copy()
        self.orig_shaded = switch.getChild(1).copy()
        self.orig_wireframe = switch.getChild(2).copy()

        # replace root node of object
        switch.replaceChild(0, flatlines)
        switch.replaceChild(1, shaded)
        switch.replaceChild(2, wireframe)

    def unloadInventor(self, obj):
        "restore original nodes"

        if (not hasattr(self, "orig_flatlines")) or (not self.orig_flatlines):
            return
        if (not hasattr(self, "orig_shaded")) or (not self.orig_shaded):
            return
        if (not hasattr(self, "orig_wireframe")) or (not self.orig_wireframe):
            return
        from pivy import coin
        from draftutils import gui_utils

        # check node contents
        rootnode = obj.ViewObject.RootNode
        # display mode switch
        switch = gui_utils.find_coin_node(obj.ViewObject.RootNode, coin.SoSwitch)
        if switch is None or switch.getNumChildren() != 4:
            FreeCAD.Console.PrintError(
                translate("Arch", "Invalid root node in") + " " + obj.Label + "\n"
            )
            return

        # replace root node of object
        switch.replaceChild(0, self.orig_flatlines)
        switch.replaceChild(1, self.orig_shaded)
        switch.replaceChild(2, self.orig_wireframe)

        # discard old content
        self.orig_flatlines = None
        self.orig_shaded = None
        self.orig_wireframe = None

    def getInventorString(self, obj):
        "locates and loads an iv file saved together with an object, if existing"

        filename = obj.Proxy.getFile(obj)
        if not filename:
            return None
        part = obj.Part
        if not obj.Part:
            return None
        zdoc = zipfile.ZipFile(filename)
        if not "Document.xml" in zdoc.namelist():
            return None
        ivfile = None
        with zdoc.open("Document.xml") as docf:
            writemode1 = False
            writemode2 = False
            for line in docf:
                line = line.decode("utf8")
                if ("<Object name=" in line) and (part in line):
                    writemode1 = True
                elif writemode1 and ('<Property name="SavedInventor"' in line):
                    writemode1 = False
                    writemode2 = True
                elif writemode2 and ("<FileIncluded file=" in line):
                    n = re.findall(r"file=\"(.*?)\"", line)
                    if n:
                        ivfile = n[0]
                        break
        if not ivfile:
            return None
        if not ivfile in zdoc.namelist():
            return None
        f = zdoc.open(ivfile)
        buf = f.read()
        buf = buf.decode("utf8")
        f.close()
        buf = buf.replace("lineWidth 2", "lineWidth " + str(int(obj.ViewObject.LineWidth)))
        return buf

    def setIFCNode(self, obj, filename):
        """Sets the coin node of this object from an IFC file"""

        try:
            from nativeifc import ifc_tools
            from nativeifc import ifc_generator
        except:
            t = translate("Arch", "NativeIFC not available - unable to process IFC files")
            FreeCAD.Console.PrintError(t + "\n")
            return
        ifcfile = obj.Proxy.getIfcFile(filename)
        elements = obj.Proxy.getIFCElements(obj, ifcfile)
        node, placement = ifc_generator.generate_coin(ifcfile, elements, cached=True)
        if node:
            ifc_generator.set_representation(obj.ViewObject, node)
            colors = node[0]
            if colors:
                ifc_tools.set_colors(obj, colors)
        else:
            ifc_generator.set_representation(obj.ViewObject, None)
        if placement:
            obj.Placement = placement


class ArchReferenceTaskPanel:
    """The editmode TaskPanel for Reference objects"""

    def __init__(self, obj):

        self.obj = obj
        self.filename = None
        self.form = QtGui.QWidget()
        self.form.setWindowTitle(translate("Arch", "External reference"))
        layout = QtGui.QVBoxLayout(self.form)
        label1 = QtGui.QLabel(translate("Arch", "External file") + ":")
        layout.addWidget(label1)
        self.fileButton = QtGui.QPushButton(self.form)
        self.openButton = QtGui.QPushButton(self.form)
        self.openButton.setText(translate("Arch", "Open"))
        if not self.obj.File:
            self.openButton.setEnabled(False)
        l2 = QtGui.QHBoxLayout()
        layout.addLayout(l2)
        l2.addWidget(self.fileButton)
        l2.addWidget(self.openButton)
        label2 = QtGui.QLabel(translate("Arch", "Part to use:"))
        layout.addWidget(label2)
        if self.obj.File:
            self.fileButton.setText(os.path.basename(self.obj.File))
        else:
            self.fileButton.setText(translate("Arch", "Choose File"))
        self.partCombo = QtGui.QComboBox(self.form)
        self.partCombo.setEnabled(False)
        layout.addWidget(self.partCombo)
        if hasattr(self.obj.Proxy, "parts"):
            parts = self.obj.Proxy.parts
        else:
            parts = self.obj.Proxy.getPartsList(self.obj)
        if parts:
            self.partCombo.setEnabled(True)
            sortedkeys = sorted(parts)
            self.partCombo.addItem(translate("Arch", "None (Use whole object)"), "")
            for k in sortedkeys:
                self.partCombo.addItem(parts[k][0], k)
            if self.obj.Part:
                if self.obj.Part in sortedkeys:
                    self.partCombo.setCurrentIndex(sortedkeys.index(self.obj.Part) + 1)
        else:
            self.partCombo.setEnabled(False)
        QtCore.QObject.connect(self.fileButton, QtCore.SIGNAL("clicked()"), self.chooseFile)
        QtCore.QObject.connect(self.openButton, QtCore.SIGNAL("clicked()"), self.openFile)

    def accept(self):

        from PySide import QtCore

        if self.filename:
            if self.filename != self.obj.File:
                self.obj.File = self.filename
                FreeCAD.ActiveDocument.recompute()
        if self.partCombo.currentText():
            i = self.partCombo.currentIndex()
            if i >= 1:
                if self.partCombo.itemData(i) != self.obj.Part:
                    self.obj.Part = self.partCombo.itemData(i)
            else:
                self.obj.Part = ""
            QtCore.QTimer.singleShot(0, FreeCAD.ActiveDocument.recompute)
        if self.filename and self.obj.Label == "External Reference":
            self.obj.Label = os.path.basename(self.filename)
        FreeCADGui.ActiveDocument.resetEdit()
        return True

    def reject(self):

        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.ActiveDocument.resetEdit()
        return True

    def chooseFile(self):

        loc = QtCore.QDir.homePath()
        if self.obj.File:
            loc = os.path.dirname(self.obj.File)
        filters = "*.FCStd *.dxf"
        # enable IFC support if NativeIFC is present
        try:
            from nativeifc import ifc_tools
        except:
            pass
        else:
            filters += " *.ifc"
        filters = translate("Arch", "Reference files") + " (" + filters + ")"
        f = QtGui.QFileDialog.getOpenFileName(
            self.form, translate("Arch", "Choose reference file"), loc, filters
        )
        if f:
            self.filename = f[0]
            self.fileButton.setText(os.path.basename(self.filename))
            parts = self.obj.Proxy.getPartsList(self.obj, self.filename)
            self.partCombo.clear()
            if parts:
                self.partCombo.setEnabled(True)
                sortedkeys = sorted(parts)
                self.partCombo.addItem(translate("Arch", "None (Use whole object)"), "")
                for k in sortedkeys:
                    self.partCombo.addItem(parts[k][0], k)
                if self.obj.Part:
                    if self.obj.Part in sortedkeys:
                        self.partCombo.setCurrentIndex(sortedkeys.index(self.obj.Part) + 1)
            else:
                self.partCombo.setEnabled(False)

    def openFile(self):

        if self.obj.File:
            if self.obj.File.lower().endswith(".fcstd"):
                FreeCAD.openDocument(self.obj.File)
            else:
                FreeCAD.loadFile(self.obj.File)
            FreeCADGui.Control.closeDialog()
            FreeCADGui.ActiveDocument.resetEdit()



# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver Elmer equation base object"
__author__ = "Markus Hovorka"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

import FreeCAD as App
from ... import equationbase
from femtools import membertools

if App.GuiUp:
    import FreeCADGui as Gui
    from femguiutils import selection_widgets


class Proxy(equationbase.BaseProxy):

    def __init__(self, obj):
        super().__init__(obj)
        obj.addProperty(
            "App::PropertyInteger",
            "Priority",
            "Base",
            (
                "Number of your choice\n"
                "The equation with highest number\n"
                "will be solved first."
            ),
            locked=True,
        )


class ViewProxy(equationbase.BaseViewProxy):

    def setEdit(self, vobj, mode=0):
        task = _TaskPanel(vobj.Object)
        Gui.Control.showDialog(task)

    def unsetEdit(self, vobj, mode=0):
        Gui.Control.closeDialog()

    def doubleClicked(self, vobj):
        if Gui.Control.activeDialog():
            Gui.Control.closeDialog()
        vobj.Document.setEdit(vobj.Object.Name)
        return True

    def getTaskWidget(self, vobj):
        return None


class _TaskPanel:

    def __init__(self, obj):
        self._obj = obj
        self._selectionWidget = selection_widgets.GeometryElementsSelection(
            obj.References, ["Solid", "Face"], False, True
        )
        # start in solid selection mode
        self._selectionWidget.rb_solid.setChecked(True)
        propWidget = obj.ViewObject.Proxy.getTaskWidget(obj.ViewObject)
        if propWidget is None:
            self.form = self._selectionWidget
        else:
            self.form = [self._selectionWidget, propWidget]
        analysis = obj.getParentGroup()
        self._mesh = membertools.get_single_member(analysis, "Fem::FemMeshObject")
        self._part = self._mesh.Shape if self._mesh is not None else None
        self._partVisible = None
        self._meshVisible = None

    def open(self):
        if self._mesh is not None and self._part is not None:
            self._meshVisible = self._mesh.ViewObject.isVisible()
            self._partVisible = self._part.ViewObject.isVisible()
            self._mesh.ViewObject.hide()
            self._part.ViewObject.show()

    def reject(self):
        self._selectionWidget.finish_selection()
        self._recomputeAndRestore()
        return True

    def accept(self):
        if self._obj.References != self._selectionWidget.references:
            self._obj.References = self._selectionWidget.references
        self._selectionWidget.finish_selection()
        self._recomputeAndRestore()
        return True

    def _restoreVisibility(self):
        if self._mesh is not None and self._part is not None:
            if self._meshVisible:
                self._mesh.ViewObject.show()
            else:
                self._mesh.ViewObject.hide()
            if self._partVisible:
                self._part.ViewObject.show()
            else:
                self._part.ViewObject.hide()

    def _recomputeAndRestore(self):
        doc = Gui.getDocument(self._obj.Document)
        doc.Document.recompute()
        self._restoreVisibility()
        # TODO: test if there is an active selection observer
        # if yes Gui.Selection.removeObserver is your friend
        doc.resetEdit()


##  @}



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix constraint fluidsection"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"


import codecs
import os
from os.path import join

import FreeCAD

from femmesh import meshtools


# ********************************************************************************************
# handle elements for constraints fluidsection with Liquid Inlet or Outlet
# belongs to write_constraints_fluidsection, should be next method
# leave the constraints fluidsection code as the last constraint method in this module
# as it is none standard constraint method compared to all other constraints
def handle_fluidsection_liquid_inlet_outlet(inpfile, ccxwriter):

    if not ccxwriter.member.geos_fluidsection:
        return inpfile

    # Fluid sections:
    # fluidsection Liquid inlet outlet objs  requires special element definition
    # to fill ccxwriter.FluidInletoutlet_ele list the mat_geo_sets are needed
    # thus this has to be after the creation of mat_geo_sets
    # different pipe cross sections will generate mat_geo_sets

    ccxwriter.FluidInletoutlet_ele = []
    ccxwriter.fluid_inout_nodes_file = join(
        ccxwriter.dir_name, f"{ccxwriter.mesh_name}_inout_nodes.txt"
    )

    def get_fluidsection_inoutlet_obj_if_setdata(matgeoset):
        if (
            matgeoset["ccx_elset"]
            and not isinstance(matgeoset["ccx_elset"], str)
            and "fluidsection_obj" in matgeoset  # fluid mesh
        ):
            fluidsec_obj = matgeoset["fluidsection_obj"]
            if fluidsec_obj.SectionType == "Liquid" and (
                fluidsec_obj.LiquidSectionType == "PIPE INLET"
                or fluidsec_obj.LiquidSectionType == "PIPE OUTLET"
            ):
                return fluidsec_obj
        return None

    def is_fluidsection_inoutlet_setnames_possible(mat_geo_sets):
        for matgeoset in mat_geo_sets:
            if matgeoset["ccx_elset"] and "fluidsection_obj" in matgeoset:  # fluid mesh
                fluidsec_obj = matgeoset["fluidsection_obj"]
                if fluidsec_obj.SectionType == "Liquid" and (
                    fluidsec_obj.LiquidSectionType == "PIPE INLET"
                    or fluidsec_obj.LiquidSectionType == "PIPE OUTLET"
                ):
                    return True
        return False

    # collect elementIDs for fluidsection Liquid inlet outlet objs
    # if they have element data (happens if not "eall")
    for matgeoset in ccxwriter.mat_geo_sets:
        fluidsec_obj = get_fluidsection_inoutlet_obj_if_setdata(matgeoset)
        if fluidsec_obj is None:
            continue
        elsetchanged = False
        counter = 0
        for elid in matgeoset["ccx_elset"]:
            counter = counter + 1
            if (elsetchanged is False) and (fluidsec_obj.LiquidSectionType == "PIPE INLET"):
                # 3rd index is to track which line nr the element is defined
                ccxwriter.FluidInletoutlet_ele.append(
                    [str(elid), fluidsec_obj.LiquidSectionType, 0]
                )
                elsetchanged = True
            elif (fluidsec_obj.LiquidSectionType == "PIPE OUTLET") and (
                counter == len(matgeoset["ccx_elset"])
            ):
                # 3rd index is to track which line nr the element is defined
                ccxwriter.FluidInletoutlet_ele.append(
                    [str(elid), fluidsec_obj.LiquidSectionType, 0]
                )

    # create the correct element definition for fluidsection Liquid inlet outlet objs
    # at least one "fluidsection_obj" needs to be in mat_geo_sets and has the attributes
    # TODO: what if there are other objs in elsets?
    if is_fluidsection_inoutlet_setnames_possible(ccxwriter.mat_geo_sets) is not None:
        # it is not distinguished if split input file
        # for split input file the main file is just closed and reopend even if not needed
        inpfile.close()
        meshtools.use_correct_fluidinout_ele_def(
            ccxwriter.FluidInletoutlet_ele, ccxwriter.femmesh_file, ccxwriter.fluid_inout_nodes_file
        )
        inpfile = codecs.open(ccxwriter.file_name, "a", encoding="utf-8")

    return inpfile


# ********************************************************************************************
# TODO:
# split method into separate methods and move some part into base writer
# see also method handle_fluidsection_liquid_inlet_outlet
def write_constraints_fluidsection(f, ccxwriter):
    if not ccxwriter.member.geos_fluidsection:
        return
    if ccxwriter.analysis_type not in ["thermomech"]:
        return

    # write constraint to file
    f.write("\n***********************************************************\n")
    f.write("** FluidSection constraints\n")
    if os.path.exists(ccxwriter.fluid_inout_nodes_file):
        inout_nodes_file = open(ccxwriter.fluid_inout_nodes_file)
        lines = inout_nodes_file.readlines()
        inout_nodes_file.close()
    else:
        FreeCAD.Console.PrintError(
            f"1DFlow inout nodes file not found: {ccxwriter.fluid_inout_nodes_file}\n"
        )
    # get nodes
    ccxwriter.get_constraints_fluidsection_nodes()
    for femobj in ccxwriter.member.geos_fluidsection:
        # femobj --> dict, FreeCAD document object is femobj["Object"]
        fluidsection_obj = femobj["Object"]
        f.write("** " + fluidsection_obj.Label + "\n")
        if fluidsection_obj.SectionType == "Liquid":
            if fluidsection_obj.LiquidSectionType == "PIPE INLET":
                f.write("**Fluid Section Inlet \n")
                if fluidsection_obj.InletPressureActive is True:
                    f.write("*BOUNDARY \n")
                    for n in femobj["Nodes"]:
                        for line in lines:
                            b = line.split(",")
                            if int(b[0]) == n and b[3] == "PIPE INLET\n":
                                # degree of freedom 2 is for defining pressure
                                f.write(
                                    "{},{},{},{}\n".format(
                                        b[0], "2", "2", fluidsection_obj.InletPressure
                                    )
                                )
                if fluidsection_obj.InletFlowRateActive is True:
                    f.write("*BOUNDARY,MASS FLOW \n")
                    for n in femobj["Nodes"]:
                        for line in lines:
                            b = line.split(",")
                            if int(b[0]) == n and b[3] == "PIPE INLET\n":
                                # degree of freedom 1 is for defining flow rate
                                # factor applied to convert unit from kg/s to t/s
                                f.write(
                                    "{},{},{},{}\n".format(
                                        b[1], "1", "1", fluidsection_obj.InletFlowRate * 0.001
                                    )
                                )
            elif fluidsection_obj.LiquidSectionType == "PIPE OUTLET":
                f.write("**Fluid Section Outlet \n")
                if fluidsection_obj.OutletPressureActive is True:
                    f.write("*BOUNDARY \n")
                    for n in femobj["Nodes"]:
                        for line in lines:
                            b = line.split(",")
                            if int(b[0]) == n and b[3] == "PIPE OUTLET\n":
                                # degree of freedom 2 is for defining pressure
                                f.write(
                                    "{},{},{},{}\n".format(
                                        b[0], "2", "2", fluidsection_obj.OutletPressure
                                    )
                                )
                if fluidsection_obj.OutletFlowRateActive is True:
                    f.write("*BOUNDARY,MASS FLOW \n")
                    for n in femobj["Nodes"]:
                        for line in lines:
                            b = line.split(",")
                            if int(b[0]) == n and b[3] == "PIPE OUTLET\n":
                                # degree of freedom 1 is for defining flow rate
                                # factor applied to convert unit from kg/s to t/s
                                f.write(
                                    "{},{},{},{}\n".format(
                                        b[1], "1", "1", fluidsection_obj.OutletFlowRate * 0.001
                                    )
                                )



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009 Yorik van Havre <yorik@uncreated.net>                        *
# *   (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides utility functions that wrap around the Qt translate functions.

This module contains auxiliary functions to translate strings
using the QtCore module.
"""
## @package translate
# \ingroup draftutils
# \brief Provides utility functions that wrap around the Qt translate function.

## \addtogroup draftutils
# @{
import PySide.QtCore as QtCore
import PySide.QtWidgets as QtWidgets

Qtranslate = QtCore.QCoreApplication.translate

try:
    _encoding = QtWidgets.QApplication.UnicodeUTF8
except AttributeError:
    _encoding = None


def translate(context, text, comment=None):
    r"""Translate the text using the Qt translate function.

    It wraps around `QtWidgets.QApplication.translate`,
    which is the same as `QtCore.QCoreApplication.translate`.

    Parameters
    ----------
    context: str
        In C++ it is typically a class name.
        But it can also be any other string to categorize the translation,
        for example, the name of a workbench, tool, or function
        that is being translated. Usually it will be the name
        of the workbench.

    text: str
        Text that will be translated. It could be a single word,
        a full sentence, paragraph, or multiple paragraphs with new lines.
        Usually the last endline character '\n'
        that finishes the string doesn't need to be included
        for translation.

    Returns
    -------
    str
        A unicode string returned by `QtWidgets.QApplication.translate`.

    Unicode strings
    ---------------
    Reference: https://doc.qt.io/qtforpython/PySide2/QtCore

    In Qt5 the strings are always assumed unicode

    >>> QtCore.QCoreApplication.translate(context, text, None)
    >>> QtWidgets.QApplication.translate(context, text, None)
    """
    return Qtranslate(context, text, comment)


# Original code no longer used. It is listed here for reference
# to show how the different pairings Py2/Qt4, Py3/Qt5, Py2/Qt5, Py3/Qt4
# were handled in the past.
# If there is a problem with the code above, this code can be made active
# again, and the code above can be commented out.
#
# =============================================================================
# try:
#     _encoding = QtWidgets.QApplication.UnicodeUTF8 if six.PY2 else None
#     def translate(context, text, utf8_decode=True):
#         """convenience function for Qt translator
#             context: str
#                 context is typically a class name (e.g., "MyDialog")
#             text: str
#                 text which gets translated
#             utf8_decode: bool [False]
#                 if set to true utf8 encoded unicode will be returned.
#                 This option does not have influence
#                 on python3 as for python3 we are returning utf-8 encoded
#                 unicode by default!
#         """
#         if six.PY3:
#             return Qtranslate(context, text, None)
#         elif utf8_decode:
#             return Qtranslate(context, text, None, _encoding)
#         else:
#             return Qtranslate(context, text, None, _encoding).encode("utf8")
#
# except AttributeError:
#     def translate(context, text, utf8_decode=False):
#         """convenience function for Qt translator
#             context: str
#                 context is typically a class name (e.g., "MyDialog")
#             text: str
#                 text which gets translated
#             utf8_decode: bool [False]
#                 if set to true utf8 encoded unicode will be returned.
#                 This option does not have influence
#                 on python3 as for python3 we are returning utf-8 encoded
#                 unicode by default!
#         """
#         if six.PY3:
#             return Qtranslate(context, text, None)
#         elif QtCore.qVersion() > "4":
#             if utf8_decode:
#                 return Qtranslate(context, text, None)
#             else:
#                 return Qtranslate(context, text, None).encode("utf8")
#         else:
#             if utf8_decode:
#                 return Qtranslate(context, text, None, _encoding)
#             else:
#                 return Qtranslate(context, text, None,
#                                   _encoding).encode("utf8")
# =============================================================================

# The same Qt translate function is provided here
QT_TRANSLATE_NOOP = QtCore.QT_TRANSLATE_NOOP


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

import sys
from lazy_loader.lazy_loader import LazyLoader
from . import toolbit
from .assets import DummyAssetSerializer
from .camassets import cam_assets
from .library import Library
from .library.serializers import FCTLSerializer
from .toolbit import ToolBit
from .toolbit.serializers import FCTBSerializer
from .shape import ToolBitShape, ToolBitShapePngIcon, ToolBitShapeSvgIcon
from .machine import Machine

# Register asset classes and serializers.
cam_assets.register_asset(Library, FCTLSerializer)
cam_assets.register_asset(ToolBit, FCTBSerializer)
cam_assets.register_asset(ToolBitShape, DummyAssetSerializer)
cam_assets.register_asset(ToolBitShapePngIcon, DummyAssetSerializer)
cam_assets.register_asset(ToolBitShapeSvgIcon, DummyAssetSerializer)
cam_assets.register_asset(Machine, DummyAssetSerializer)

# For backward compatibility with files saved before the toolbit rename
# This makes the Path.Tool.toolbit.base module available as Path.Tool.Bit.
# Since C++ does not use the standard Python import mechanism and instead
# unpickles existing objects after looking them up in sys.modules, we
# need to update sys.modules here.
sys.modules[__name__ + ".Bit"] = toolbit.models.base
sys.modules[__name__ + ".Gui.Bit"] = LazyLoader(
    "Path.Tool.toolbit.ui.view", globals(), "Path.Tool.toolbit.ui.view"
)

# Define __all__ for explicit public interface
__all__ = [
    "ToolBit",
    "cam_assets",
]



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
from __future__ import annotations
import urllib.parse
from typing import Dict, Any, Mapping


class AssetUri:
    """
    Represents an asset URI with components.

    The URI structure is: <asset_type>://<asset_id>[/version]
    """

    def __init__(self, uri_string: str):
        # Manually parse the URI string
        parts = uri_string.split("://", 1)
        if len(parts) != 2:
            raise ValueError(f"Invalid URI structure: {uri_string}")

        self.asset_type = parts[0]
        rest = parts[1]

        # Split asset_id, version, and params
        path_and_query = rest.split("?", 1)
        path_parts = path_and_query[0].split("/")

        if not path_parts or not path_parts[0]:
            raise ValueError(f"Invalid URI structure: {uri_string}")

        self.asset_id = path_parts[0]
        self.version = path_parts[1] if len(path_parts) > 1 else None

        if len(path_parts) > 2:
            raise ValueError(f"Invalid URI path structure: {uri_string}")

        self.params: Dict[str, list[str]] = {}
        if len(path_and_query) > 1:
            self.params = urllib.parse.parse_qs(path_and_query[1])

        if not self.asset_type or not self.asset_id:
            raise ValueError(f"Invalid URI structure: {uri_string}")

    def __str__(self) -> str:
        path = f"/{self.version}" if self.version else ""

        query = urllib.parse.urlencode(self.params, doseq=True) if self.params else ""

        uri_string = urllib.parse.urlunparse((self.asset_type, self.asset_id, path, "", query, ""))
        return uri_string

    def __repr__(self) -> str:
        return f"AssetUri('{str(self)}')"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, AssetUri):
            return NotImplemented
        return (
            self.asset_type == other.asset_type
            and self.asset_id == other.asset_id
            and self.version == other.version
            and self.params == other.params
        )

    def __hash__(self) -> int:
        """Returns a hash value for the AssetUri."""
        return hash((self.asset_type, self.asset_id, self.version, frozenset(self.params.items())))

    @classmethod
    def is_uri(cls, uri: AssetUri | str) -> bool:
        """Checks if the given string is a valid URI."""
        if isinstance(uri, AssetUri):
            return True

        try:
            AssetUri(uri)
        except ValueError:
            return False
        return True

    @staticmethod
    def build(
        asset_type: str,
        asset_id: str,
        version: str | None = None,
        params: Mapping[str, str | list[str]] | None = None,
    ) -> AssetUri:
        """Builds a Uri object from components."""
        uri = AssetUri.__new__(AssetUri)  # Create a new instance without calling __init__
        uri.asset_type = asset_type
        uri.asset_id = asset_id
        uri.version = version
        uri.params = {}
        if params:
            for key, value in params.items():
                if isinstance(value, list):
                    uri.params[key] = value
                else:
                    uri.params[key] = [value]
        return uri



# ***************************************************************************
# *   Copyright (c) 2022 Ajinkya Dahale <dahale.a.p@gmail.com>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix constraint rigid body"
__author__ = "Ajinkya Dahale"
__url__ = "https://www.freecad.org"


import FreeCAD


def get_analysis_types():
    return "all"  # write for all analysis types


def get_sets_name():
    return "constraints_rigidbody_node_sets"


def get_constraint_title():
    return "Rigid Body Constraints"


def get_before_write_meshdata_constraint():
    return ""


def get_after_write_meshdata_constraint():
    return ""


def get_before_write_constraint():
    return ""


def get_after_write_constraint():
    return ""


def write_constraint(f, femobj, rb_obj, ccxwriter):

    rb_obj_idx = ccxwriter.analysis.Group.index(rb_obj)
    node_count = ccxwriter.mesh_object.FemMesh.NodeCount
    # factor 2 is to prevent conflict with other rigid body constraint
    ref_node_idx = node_count + 2 * rb_obj_idx + 1
    rot_node_idx = node_count + 2 * rb_obj_idx + 2

    if rb_obj.EnableAmplitude:
        rb_amplitude = f", AMPLITUDE={rb_obj.Name}"
    else:
        rb_amplitude = ""

    def write_mode(mode, node, dof, constraint, load):
        if mode == "Constraint":
            f.write(f"*BOUNDARY{rb_amplitude}\n")
            f.write(f"{node},{dof},{dof},{constraint:.13G}\n")
        elif mode == "Load":
            f.write(f"*CLOAD{rb_amplitude}\n")
            f.write(f"{node},{dof},{load:.13G}\n")

    mode = [rb_obj.TranslationalModeX, rb_obj.TranslationalModeY, rb_obj.TranslationalModeZ]
    constraint = rb_obj.Displacement
    load = [rb_obj.ForceX, rb_obj.ForceY, rb_obj.ForceZ]

    for i in range(3):
        write_mode(mode[i], ref_node_idx, i + 1, constraint[i], load[i].getValueAs("N").Value)

    mode = [rb_obj.RotationalModeX, rb_obj.RotationalModeY, rb_obj.RotationalModeZ]
    load = [rb_obj.MomentX, rb_obj.MomentY, rb_obj.MomentZ]

    # write rotation components according to rotational mode
    rot = rb_obj.Rotation
    proj_axis = [rot.Axis[i] if mode[i] == "Constraint" else 0 for i in range(3)]
    # proj_axis could be null
    try:
        constraint = FreeCAD.Vector(proj_axis).normalize() * rot.Angle
    except:
        constraint = FreeCAD.Vector(0, 0, 0)

    for i in range(3):
        write_mode(mode[i], rot_node_idx, i + 1, constraint[i], load[i].getValueAs("N*mm").Value)

    f.write("\n")



# ***************************************************************************
# *   Copyright (c) 2016 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver writer base object"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

import os
from os.path import join

import FreeCAD

from femmesh import meshsetsgetter


class FemInputWriter:
    def __init__(
        self, analysis_obj, solver_obj, mesh_obj, member, dir_name=None, mat_geo_sets=None
    ):
        # class attributes from parameter values
        self.analysis = analysis_obj
        self.solver_obj = solver_obj
        self.mesh_object = mesh_obj
        self.member = member
        # more attributes
        self.analysis_type = self.solver_obj.AnalysisType
        self.document = self.analysis.Document
        # working dir
        # if dir_name was not given or if it exists but is not empty: create a temporary dir
        # Purpose: makes sure the analysis can be run even on wired situation
        make_tmp_dir = False
        if dir_name is None:
            FreeCAD.Console.PrintWarning(
                "Error: The working_dir in base input file writer class was not set. "
                "A temporary directory is used.\n"
            )
            make_tmp_dir = True
        elif not os.path.isdir(dir_name):
            FreeCAD.Console.PrintWarning(
                "Error: The working_dir: '{}' given to "
                "base input file writer class does not exist. "
                "A temporary directory is used.\n".format(dir_name)
            )
            make_tmp_dir = True
        if make_tmp_dir is True:
            from tempfile import mkdtemp

            dir_name = mkdtemp(prefix="fcfem_")
            FreeCAD.Console.PrintWarning(
                f"The working directory '{dir_name}' was created and will be used."
            )
        self.dir_name = dir_name

        # new class attributes
        self.fc_ver = FreeCAD.Version()
        self.ccx_nall = "Nall"
        self.ccx_eall = "Eall"
        self.ccx_evolumes = "Evolumes"
        self.ccx_efaces = "Efaces"
        self.ccx_eedges = "Eedges"
        self.mat_geo_sets = mat_geo_sets
        if self.mesh_object:
            self.femmesh = self.mesh_object.FemMesh
        else:
            FreeCAD.Console.PrintWarning(
                "No finite element mesh object was given to the writer class. "
                "In rare cases this might not be an error. "
            )

        # *************************************************
        # deprecated, leave for compatibility reasons
        # if these are calculated here they are calculated twice :-(
        self.femnodes_mesh = {}
        self.femelement_table = {}
        self.constraint_conflict_nodes = []
        self.femnodes_ele_table = {}
        self.femelements_edges_only = []
        self.femelements_faces_only = []
        self.femelement_volumes_table = {}
        self.femelement_faces_table = {}
        self.femelement_edges_table = {}
        self.femelement_count_test = True

        # deprecated, leave for compatibility reasons
        # do not add new objects
        # only the ones which exists on 0.19 release are kept
        # materials
        self.material_objects = member.mats_linear
        self.material_nonlinear_objects = member.mats_nonlinear
        # geometries
        self.beamsection_objects = member.geos_beamsection
        self.beamrotation_objects = member.geos_beamrotation
        self.fluidsection_objects = member.geos_fluidsection
        self.shellthickness_objects = member.geos_shellthickness
        # constraints
        self.contact_objects = member.cons_contact
        self.displacement_objects = member.cons_displacement
        self.fixed_objects = member.cons_fixed
        self.force_objects = member.cons_force
        self.heatflux_objects = member.cons_heatflux
        self.initialtemperature_objects = member.cons_initialtemperature
        self.planerotation_objects = member.cons_planerotation
        self.pressure_objects = member.cons_pressure
        self.selfweight_objects = member.cons_selfweight
        self.temperature_objects = member.cons_temperature
        self.tie_objects = member.cons_tie
        self.transform_objects = member.cons_transform
        self.electrostatic_objects = member.cons_electrostatic
        self.electricchargedensity_objects = member.cons_electricchargedensity

        # meshdatagetter, for compatibility, same with all getter methods
        self.meshdatagetter = meshsetsgetter.MeshSetsGetter(
            self.analysis,
            self.solver_obj,
            self.mesh_object,
            self.member,
        )

    # ********************************************************************************************
    # ********************************************************************************************
    # generic writer for constraints mesh sets and constraints property data
    # write constraint node sets, constraint face sets, constraint element sets
    def write_constraints_meshsets(self, f, femobjs, con_module):
        if not femobjs:
            return

        analysis_types = con_module.get_analysis_types()
        if analysis_types != "all" and self.analysis_type not in analysis_types:
            return

        def constraint_sets_loop_writing(the_file, femobjs, write_before, write_after):
            if write_before != "":
                the_file.write(write_before)
            for femobj in femobjs:
                # femobj --> dict, FreeCAD document object is femobj["Object"]
                the_obj = femobj["Object"]
                the_file.write(f"** {the_obj.Label}\n")
                con_module.write_meshdata_constraint(the_file, femobj, the_obj, self)
            if write_after != "":
                the_file.write(write_after)

        write_before = con_module.get_before_write_meshdata_constraint()
        write_after = con_module.get_after_write_meshdata_constraint()

        # write sets to file
        write_name = con_module.get_sets_name()
        f.write("\n{}\n".format(59 * "*"))
        f.write("** {}\n".format(write_name.replace("_", " ")))

        if self.split_inpfile is True:
            file_name_split = f"{self.mesh_name}_{write_name}.inp"
            f.write(f"*INCLUDE,INPUT={file_name_split}\n")
            inpfile_split = open(join(self.dir_name, file_name_split), "w")
            constraint_sets_loop_writing(inpfile_split, femobjs, write_before, write_after)
            inpfile_split.close()
        else:
            constraint_sets_loop_writing(f, femobjs, write_before, write_after)

    # write constraint property data
    def write_constraints_propdata(self, f, femobjs, con_module):

        if not femobjs:
            return

        analysis_types = con_module.get_analysis_types()
        if analysis_types != "all" and self.analysis_type not in analysis_types:
            return

        write_before = con_module.get_before_write_constraint()
        write_after = con_module.get_after_write_constraint()

        # write constraint to file
        f.write("\n{}\n".format(59 * "*"))
        f.write(f"** {con_module.get_constraint_title()}\n")
        if write_before != "":
            f.write(write_before)
        for femobj in femobjs:
            # femobj --> dict, FreeCAD document object is femobj["Object"]
            the_obj = femobj["Object"]
            f.write(f"** {the_obj.Label}\n")
            con_module.write_constraint(f, femobj, the_obj, self)
        if write_after != "":
            f.write(write_after)

    # ********************************************************************************************
    # deprecated, do not add new constraints
    # only the ones which exists on 0.19 release are kept
    def get_constraints_fixed_nodes(self):
        self.meshdatagetter.get_constraints_fixed_nodes()

    def get_constraints_displacement_nodes(self):
        self.meshdatagetter.get_constraints_displacement_nodes()

    def get_constraints_planerotation_nodes(self):
        self.meshdatagetter.get_constraints_planerotation_nodes()

    def get_constraints_transform_nodes(self):
        self.meshdatagetter.get_constraints_transform_nodes()

    def get_constraints_temperature_nodes(self):
        self.meshdatagetter.get_constraints_temperature_nodes()

    def get_constraints_fluidsection_nodes(self):
        self.meshdatagetter.get_constraints_fluidsection_nodes()

    def get_constraints_force_nodeloads(self):
        self.meshdatagetter.get_constraints_force_nodeloads()

    def get_constraints_pressure_faces(self):
        self.meshdatagetter.get_constraints_pressure_faces()

    def get_constraints_contact_faces(self):
        self.meshdatagetter.get_constraints_contact_faces()

    def get_constraints_tie_faces(self):
        self.meshdatagetter.get_constraints_tie_faces()

    def get_constraints_heatflux_faces(self):
        self.meshdatagetter.get_constraints_heatflux_faces()


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import tempfile

import ifcopenshell

import FreeCAD
import Draft

from importers import exportIFC
from importers import exportIFCHelper
from importers import importIFCHelper

from . import ifc_tools
from . import ifc_import


PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/NativeIFC")


def get_export_preferences(ifcfile, preferred_context=None, create=None):
    """returns a preferences dict for exportIFC.
    Preferred context can either indicate a ContextType like 'Model' or 'Plan',
    or a [ContextIdentifier,ContextType,TargetView] list or tuple, for ex.
    ('Annotation','Plan') or ('Body','Model','MODEL_VIEW'). This function
    will do its best to find the most appropriate context. If create is True,
    if the exact context is not found, a new one is created"""

    prefs = exportIFC.getPreferences()
    prefs["SCHEMA"] = ifcfile.wrapped_data.schema_name()
    s = ifcopenshell.util.unit.calculate_unit_scale(ifcfile)
    # the above lines yields meter -> file unit scale factor. We need mm
    prefs["SCALE_FACTOR"] = 0.001 / s
    cids = ifc_tools.get_body_context_ids(ifcfile)
    contexts = [ifcfile[i] for i in cids]
    best_context = None
    exact_match = False
    if preferred_context:
        if isinstance(preferred_context, str):
            for context in contexts:
                if context.ContextType == preferred_context:
                    best_context = context
                    exact_match = True
                    break
        elif isinstance(preferred_context, (list, tuple)):
            second_choice = None
            for context in contexts:
                if len(preferred_context) > 2:
                    if (
                        context.TargetView == preferred_context[2]
                        and context.ContextType == preferred_context[1]
                        and context.ContextIdentifier == preferred_context[0]
                    ):
                        best_context = context
                        exact_match = True
                if len(preferred_context) > 1:
                    if (
                        context.ContextType == preferred_context[1]
                        and context.ContextIdentifier == preferred_context[0]
                    ):
                        if not exact_match:
                            best_context = context
                            if len(preferred_context) == 2:
                                exact_match = True
                if context.ContextType == preferred_context[0]:
                    if not exact_match:
                        best_context = context
                        if len(preferred_context) == 1:
                            exact_match = True
            if contexts:
                if not best_context:
                    best_context = contexts[0]
        if create:
            if not exact_match:
                if isinstance(preferred_context, str):
                    best_context = ifc_tools.api_run(
                        "context.add_context", ifcfile, context_type=preferred_context
                    )
                elif best_context:
                    if len(preferred_context) > 2:
                        best_context = ifc_tools.api_run(
                            "context.add_context",
                            ifcfile,
                            context_type=preferred_context[1],
                            context_identifier=preferred_context[0],
                            target_view=preferred_context[2],
                            parent=best_context,
                        )
                    elif len(preferred_context) > 1:
                        best_context = ifc_tools.api_run(
                            "context.add_context",
                            ifcfile,
                            context_type=preferred_context[1],
                            context_identifier=preferred_context[0],
                            parent=best_context,
                        )
                else:
                    if len(preferred_context) > 1:
                        best_context = ifc_tools.api_run(
                            "context.add_context", ifcfile, context_type=preferred_context[1]
                        )
                        if len(preferred_context) > 2:
                            best_context = ifc_tools.api_run(
                                "context.add_context",
                                ifcfile,
                                context_type=preferred_context[1],
                                context_identifier=preferred_context[0],
                                target_view=preferred_context[2],
                                parent=best_context,
                            )
                        else:
                            best_context = ifc_tools.api_run(
                                "context.add_context",
                                ifcfile,
                                context_type=preferred_context[1],
                                context_identifier=preferred_context[0],
                                parent=best_context,
                            )
                    else:
                        best_context = ifc_tools.api_run(
                            "context.add_context", ifcfile, context_type=preferred_context[0]
                        )
    if not best_context:
        if contexts:
            best_context = contexts[0]
        else:
            best_context = ifc_tools.api_run("context.add_context", ifcfile, context_type="Model")
    return prefs, best_context


def create_product(obj, parent, ifcfile, ifcclass=None):
    """Creates an IFC product out of a FreeCAD object"""

    name = obj.Label
    description = getattr(obj, "Description", None)
    if not ifcclass:
        ifcclass = ifc_tools.get_ifctype(obj)
    representation, placement = create_representation(obj, ifcfile)
    product = ifc_tools.api_run("root.create_entity", ifcfile, ifc_class=ifcclass, name=name)
    ifc_tools.set_attribute(ifcfile, product, "Description", description)
    ifc_tools.set_attribute(ifcfile, product, "ObjectPlacement", placement)
    # TODO below cannot be used at the moment because the ArchIFC exporter returns an
    # IfcProductDefinitionShape already and not an IfcShapeRepresentation
    # ifc_tools.api_run("geometry.assign_representation", ifcfile, product=product, representation=representation)
    ifc_tools.set_attribute(ifcfile, product, "Representation", representation)
    # TODO treat subtractions/additions
    return product


def create_representation(obj, ifcfile):
    """Creates a geometry representation for the given object"""

    # TEMPORARY use the Arch exporter
    # TODO this is temporary. We should rely on ifcopenshell for this with:
    # https://blenderbim.org/docs-python/autoapi/ifcopenshell/api/root/create_entity/index.html
    # a new FreeCAD 'engine' should be added to:
    # https://blenderbim.org/docs-python/autoapi/ifcopenshell/api/geometry/index.html
    # that should contain all typical use cases one could have to convert FreeCAD geometry
    # to IFC.

    # setup exporter - TODO do that in the module init
    exportIFC.clones = {}
    exportIFC.profiledefs = {}
    exportIFC.surfstyles = {}
    exportIFC.shapedefs = {}
    exportIFC.ifcopenshell = ifcopenshell
    exportIFC.ifcbin = exportIFCHelper.recycler(ifcfile, template=False)
    prefs, context = get_export_preferences(ifcfile)
    representation, placement, shapetype = exportIFC.getRepresentation(
        ifcfile, context, obj, preferences=prefs
    )
    return representation, placement


def get_object_type(ifcentity, objecttype=None):
    """Determines a creation type for this object"""

    if not objecttype:
        if ifcentity.is_a("IfcAnnotation"):
            if get_sectionplane(ifcentity):
                objecttype = "sectionplane"
            elif get_dimension(ifcentity):
                objecttype = "dimension"
            elif get_text(ifcentity):
                objecttype = "text"
        elif ifcentity.is_a("IfcGridAxis"):
            objecttype = "axis"
        elif ifcentity.is_a("IfcControl"):
            objecttype = "schedule"
        elif ifcentity.is_a() in ["IfcBuilding", "IfcBuildingStorey"]:
            objecttype = "buildingpart"
    return objecttype


def is_annotation(obj):
    """Determines if the given FreeCAD object should be saved as an IfcAnnotation"""

    if getattr(obj, "IfcClass", None) in ["IfcAnnotation", "IfcGridAxis"]:
        return True
    if getattr(obj, "IfcType", None) == "Annotation":
        return True
    if obj.isDerivedFrom("Part::Part2DObject"):
        return True
    elif obj.isDerivedFrom("App::Annotation"):
        return True
    elif Draft.getType(obj) in [
        "BezCurve",
        "BSpline",
        "Wire",
        "DraftText",
        "Text",
        "Dimension",
        "LinearDimension",
        "AngularDimension",
        "SectionPlane",
    ]:
        return True
    elif obj.isDerivedFrom("Part::Feature"):
        if obj.Shape and (not obj.Shape.Solids) and obj.Shape.Edges:
            if not obj.Shape.Faces:
                return True
            elif (
                (obj.Shape.BoundBox.XLength < 0.0001)
                or (obj.Shape.BoundBox.YLength < 0.0001)
                or (obj.Shape.BoundBox.ZLength < 0.0001)
            ):
                return True
    return False


def get_text(annotation):
    """Determines if an IfcAnnotation contains an IfcTextLiteral.
    Returns the IfcTextLiteral or None"""

    if annotation.is_a("IfcAnnotation"):
        for rep in annotation.Representation.Representations:
            for item in rep.Items:
                if item.is_a("IfcTextLiteral"):
                    return item
    return None


def get_dimension(annotation):
    """Determines if an IfcAnnotation is representing a dimension.
    Returns a list containing the representation, two points indicating
    the measured points, and optionally a third point indicating where
    the dimension line is located, if available"""

    if annotation.is_a("IfcAnnotation"):
        if annotation.ObjectType == "DIMENSION":
            s = ifcopenshell.util.unit.calculate_unit_scale(annotation.file) * 1000
            for rep in annotation.Representation.Representations:
                shape = importIFCHelper.get2DShape(rep, s, notext=True)
                pl = get_placement(annotation.ObjectPlacement, scale=s)
                if pl:
                    shape[0].Placement = pl
                if shape and len(shape) == 1:
                    if len(shape[0].Vertexes) >= 2:
                        # two-point polyline (BBIM)
                        res = [rep, shape[0].Vertexes[0].Point, shape[0].Vertexes[-1].Point]
                        if len(shape[0].Vertexes) > 2:
                            # 4-point polyline (FreeCAD)
                            res.append(shape[0].Vertexes[1].Point)
                        return res
                else:
                    print(annotation, "NOT A DIMENSION")
    return None


def get_sectionplane(annotation):
    """Determines if an IfcAnnotation is representing a section plane.
    Returns a list containing a placement, and optionally an X dimension,
    an Y dimension and a  depth dimension"""

    if annotation.is_a("IfcAnnotation"):
        if annotation.ObjectType == "DRAWING":
            s = ifcopenshell.util.unit.calculate_unit_scale(annotation.file) * 1000
            result = [get_placement(annotation.ObjectPlacement, scale=s)]
            for rep in annotation.Representation.Representations:
                for item in rep.Items:
                    if item.is_a("IfcCsgSolid"):
                        if item.TreeRootExpression.is_a("IfcBlock"):
                            result.append(item.TreeRootExpression.XLength * s)
                            result.append(item.TreeRootExpression.YLength * s)
                            result.append(item.TreeRootExpression.ZLength * s)
            return result
    return None


def get_axis(obj):
    """Determines if a given IFC entity is an IfcGridAxis. Returns a tuple
    containing a Placement, a length value in millimeters, and a tag"""

    if obj.is_a("IfcGridAxis"):
        tag = obj.AxisTag
        s = ifcopenshell.util.unit.calculate_unit_scale(obj.file) * 1000
        shape = importIFCHelper.get2DShape(obj.AxisCurve, s, notext=True)
        if shape:
            edge = shape[0].Edges[0]  # we suppose here the axis shape is a single straight line
            if obj.SameSense:
                p0 = edge.Vertexes[0].Point
                p1 = edge.Vertexes[-1].Point
            else:
                p0 = edge.Vertexes[-1].Point
                p1 = edge.Vertexes[0].Point
            length = edge.Length
            placement = FreeCAD.Placement()
            placement.Base = p0
            placement.Rotation = FreeCAD.Rotation(FreeCAD.Vector(0, 1, 0), p1.sub(p0))
            return (placement, length, tag)
    return None


def create_annotation(obj, ifcfile):
    """Adds an IfcAnnotation from the given object to the given IFC file"""

    exportIFC.clones = {}
    exportIFC.profiledefs = {}
    exportIFC.surfstyles = {}
    exportIFC.shapedefs = {}
    exportIFC.curvestyles = {}
    exportIFC.ifcopenshell = ifcopenshell
    exportIFC.ifcbin = exportIFCHelper.recycler(ifcfile, template=False)
    if is_annotation(obj) and Draft.getType(obj) != "SectionPlane":
        context_type = "Plan"
    else:
        context_type = "Model"
    prefs, context = get_export_preferences(ifcfile, preferred_context=context_type, create=True)
    prefs["BBIMDIMS"] = True  # Save dimensions as 2-point polylines
    history = get_history(ifcfile)
    # TODO The following prints each edge as a separate IfcGeometricCurveSet
    # It should be refined to create polylines instead
    anno = exportIFC.create_annotation(obj, ifcfile, context, history, prefs)
    return anno


def get_history(ifcfile):
    """Returns the owner history or None"""

    history = ifcfile.by_type("IfcOwnerHistory")
    if history:
        history = history[0]
    else:
        # IFC4 allows to not write any history
        history = None
    return history


def get_placement(ifcelement, ifcfile=None, scale=None):
    """Returns a FreeCAD placement from an IFC placement"""

    if not scale:
        if not ifcfile:
            ifcfile = ifcelement.file
        scale = 0.001 / ifcopenshell.util.unit.calculate_unit_scale(ifcfile)
    return importIFCHelper.getPlacement(ifcelement, scaling=scale)


def get_scaled_point(point, ifcfile=None, is2d=False):
    """Returns a scaled 2d or 3d point tuple form a FreeCAD point"""

    if not ifcfile:
        ifcfile = ifcelement.file
    s = 0.001 / ifcopenshell.util.unit.calculate_unit_scale(ifcfile)
    v = FreeCAD.Vector(point)
    v.multiply(s)
    v = tuple(v)
    if is2d:
        v = v[:2]
    return v


def get_scaled_value(value, ifcfile):
    """Returns a scaled dimension value"""

    s = 0.001 / ifcopenshell.util.unit.calculate_unit_scale(ifcfile)
    return value * s


def export_and_convert(objs, doc):
    """Exports the given objects and their descendents to the given IFC file
    and re-imports it into the given document. This is slower than direct_conversion()
    but gives an intermediate file which can be useful for debugging"""

    tf = tempfile.mkstemp(suffix=".ifc")[1]
    exportIFC.export(objs, tf)
    ifc_import.insert(tf, doc.Name, singledoc=True)


def direct_conversion(objs, doc):
    """Exports the given objects to the given ifcfile and recreates the contents"""

    prj_obj = ifc_tools.convert_document(doc, silent=True)
    exportIFC.export(objs, doc.Proxy.ifcfile)
    if PARAMS.GetBool("LoadOrphans", True):
        ifc_tools.load_orphans(prj_obj)
    if PARAMS.GetBool("LoadMaterials", False):
        ifc_materials.load_materials(prj_obj)
    if PARAMS.GetBool("LoadLayers", False):
        ifc_layers.load_layers(prj_obj)
    if PARAMS.GetBool("LoadPsets", False):
        ifc_psets.load_psets(prj_obj)



# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *   Copyright (c) 2018 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Collection of functions for the Fem module.

This module contains function for managing a analysis and all the different
types of objects it contains, helper for executing a simulation.
"""


__title__ = "FEM analysis tools"
__author__ = "Markus Hovorka, Bernd Hahnebach"
__url__ = "https://www.freecad.org"


from . import femutils


def get_member(analysis, t):
    """Return list of all members of *analysis* of type *t*.

    Search *analysis* for members of type *t*. This method checks the custom
    python typesystem (BaseType class property) used by the Fem module if
    possible. If the object does not use the python typesystem the usual
    isDerivedFrom from the C++ dynamic type system is used.

    :param analysis: only objects part of this analysis are considered
    :param t:        only objects of this type are returned

    :note:
     Inheritance of Fem types is not checked. If *obj* uses Fems typesystem the
     type is just checked for equality. If the type doesn't match
     ``obj.isDerivedFrom`` is called as usual. See
     https://forum.freecad.org/viewtopic.php?f=10&t=32625
    """
    if analysis is None:
        raise ValueError("Analysis must not be None")
    matching = []
    for m in analysis.Group:
        # since is _derived_from is used the father could be used
        # to test too (ex. "Fem::FemMeshObject")
        if femutils.is_derived_from(m, t) and not (
            m.hasExtension("App::SuppressibleExtension") and m.Suppressed
        ):
            matching.append(m)
    return matching


def get_single_member(analysis, t):
    """Return one object of type *t* and part of *analysis*.

    Search *analysis* for members of type *t* and return the first one that's
    found. This method checks the custom python typesystem (BaseType class
    property) used by the Fem module if possible. If the object doesn't use the
    python typesystem the usual isDerivedFrom from the C++ dynamic type system
    is used.

    :param analysis: only objects part of this analysis are considered
    :param t:        only a object of this type is returned

    :note:
     Inheritance of Fem types is not checked. If *obj* uses Fems typesystem the
     type is just checked for equality. If the type doesn't match
     ``obj.isDerivedFrom`` is called as usual. See
     https://forum.freecad.org/viewtopic.php?f=10&t=32625
    """
    objs = get_member(analysis, t)
    return objs[0] if objs else None


def get_several_member(analysis, t):
    """Get members and pack them for Calculix/Z88.

    Collect members by calling :py:func:`get_member` and pack them into a
    data structure that can be consumed by calculix and Z88 solver modules.

    :param analysis: see :py:func:`get_member`
    :param t: see :py:func:`get_member`

    :returns:
     A list containing one dict per member. Each dict has two entries:
     ``"Object"`` and ``"RefShapeType"``. ``dict["Object"]`` contains the
     member document object. ``dict["RefShapeType"]`` contains the shape type
     of the *References* property of the member (used by constraints) as a
     string ("Vertex", "Edge", "Face" or "Solid"). If the member doesn't have a
     *References* property ``dict["RefShapeType"]`` is the empty string ``""``.

    :note:
     Undefined behaviour if one of the members has a *References* property
     which is empty.

    :note:
     Undefined behaviour if the type of the references of one object are not
     all the same.

    :note:
     Inheritance of Fem types is not checked. If *obj* uses Fems typesystem the
     type is just checked for equality. If the type doesn't match
     ``obj.isDerivedFrom`` is called as usual. See
     https://forum.freecad.org/viewtopic.php?f=10&t=32625
    """
    # if no member is found, an empty list is returned
    objs = get_member(analysis, t)
    members = []
    for m in objs:
        obj_dict = {}
        obj_dict["Object"] = m
        obj_dict["RefShapeType"] = femutils.get_refshape_type(m)
        members.append(obj_dict)
    return members


def get_mesh_to_solve(analysis):
    """Find one and only mesh object of *analysis*.

    :returns:
     A tuple ``(object, message)``. If and only if the analysis contains
     exactly one mesh object the first value of the tuple is the mesh document
     object. Otherwise the first value is ``None`` and the second value is a
     error message indicating what went wrong.
    """
    mesh_to_solve = None
    for m in analysis.Group:
        if m.isDerivedFrom("Fem::FemMeshObject") and not m.Suppressed:
            if not mesh_to_solve:
                mesh_to_solve = m
            else:
                return (None, "FEM: multiple mesh in analysis not yet supported!")
    if mesh_to_solve is not None:
        return (mesh_to_solve, "")
    else:
        return (None, "FEM: no mesh object found in analysis.")


class AnalysisMember:

    def __init__(self, analysis):
        self.analysis = analysis
        """
        # members of the analysis. All except solvers and the mesh

        materials:
        materials_linear : list of dictionaries
            list of nonlinear materials from the analysis.
            [{"Object":materials_linear}, {}, ...]

        materials_nonlinear : list of dictionaries
            list of nonlinear materials from the analysis.
            [{"Object":materials_nonlinear}, {}, ...]

        geometries:
        beam_sections : list of dictionaries
            list of beam sections from the analysis.
            [{"Object":beam_section_obj, "xxxxxxxx":value}, {}, ...]

        beam_rotations :  list of dictionaries
            list of beam rotations from the analysis.
            [{"Object":beam_rotation_obj, "xxxxxxxx":value}, {}, ...]

        fluid_sections : list of dictionaries
            list of fluid sections from the analysis.
            [{"Object":fluid_section_obj, "xxxxxxxx":value}, {}, ...]

        shell_thicknesses : list of dictionaries
            list of shell thicknesses from the analysis.
            [{"Object":shell_thickness_obj, "xxxxxxxx":value}, {}, ...]

        constraints:
        constraints_centrif : list of dictionaries
            list of centrifs for the analysis.
            [{"Object":centrif_obj, "xxxxxxxx":value}, {}, ...]

        constraints_contact : list of dictionaries
            list of contact constraints from the analysis.
            [{"Object":contact_obj, "xxxxxxxx":value}, {}, ...]

        constraints_displacement : list of dictionaries
            list of displacements for the analysis.
            [{"Object":displacement_obj, "xxxxxxxx":value}, {}, ...]

        constraints_fixed :  list of dictionaries
            list of fixed constraints from the analysis.
            [{"Object":fixed_obj, "NodeSupports":bool}, {}, ...]

        constraints_rigidbody : list of dictionaries
            list of displacements for the analysis.
            [{"Object":rigidbody_obj, "xxxxxxxx":value}, {}, ...]

        constraints_force : list of dictionaries
            list of force constraints from the analysis.
            [{"Object":force_obj, "NodeLoad":value}, {}, ...

        constraints_heatflux : list of dictionaries
            list of heatflux constraints for the analysis.
            [{"Object":heatflux_obj, "xxxxxxxx":value}, {}, ...]

        constraints_initialtemperature : list of dictionaries
            list of initial temperatures for the analysis.
            [{"Object":initialtemperature_obj, "xxxxxxxx":value}, {}, ...]

        constraints_planerotation : list of dictionaries
            list of plane rotation constraints from the analysis.
            [{"Object":planerotation_obj, "xxxxxxxx":value}, {}, ...]

        constraints_pressure : list of dictionaries
            list of pressure constraints from the analysis.
            [{"Object":pressure_obj, "xxxxxxxx":value}, {}, ...]

        constraints_sectionprint : list of dictionaries
            list of sectionprints for the analysis.
            [{"Object":sectionprint_obj, "xxxxxxxx":value}, {}, ...]

        constraints_selfweight : list of dictionaries
            list of selfweight constraints from the analysis.
            [{"Object":selfweight_obj, "xxxxxxxx":value}, {}, ...]

        constraints_temperature : list of dictionaries
            list of temperatures for the analysis.
            [{"Object":temperature_obj, "xxxxxxxx":value}, {}, ...]

        constraints_tie : list of dictionaries
            list of ties for the analysis.
            [{"Object":tie_obj, "xxxxxxxx":value}, {}, ...]

        constraints_transform : list of dictionaries
            list of transform constraints from the analysis.
            [{"Object":transform_obj, "xxxxxxxx":value}, {}, ...]
        """

        # get member
        # constants
        self.cota_vacuumpermittivity = self.get_several_member("Fem::ConstantVacuumPermittivity")

        # materials
        std_mats = self.get_several_member("Fem::MaterialCommon")
        rei_mats = self.get_several_member("Fem::MaterialReinforced")
        self.mats_linear = std_mats + rei_mats

        self.mats_nonlinear = self.get_several_member("Fem::MaterialMechanicalNonlinear")

        # geometries
        self.geos_beamsection = self.get_several_member("Fem::ElementGeometry1D")
        self.geos_beamrotation = self.get_several_member("Fem::ElementRotation1D")
        self.geos_fluidsection = self.get_several_member("Fem::ElementFluid1D")
        self.geos_shellthickness = self.get_several_member("Fem::ElementGeometry2D")

        # constraints
        self.cons_centrif = self.get_several_member("Fem::ConstraintCentrif")
        self.cons_bodyheatsource = self.get_several_member("Fem::ConstraintBodyHeatSource")
        self.cons_contact = self.get_several_member("Fem::ConstraintContact")
        self.cons_displacement = self.get_several_member("Fem::ConstraintDisplacement")
        self.cons_fixed = self.get_several_member("Fem::ConstraintFixed")
        self.cons_rigidbody = self.get_several_member("Fem::ConstraintRigidBody")
        self.cons_rigidbody_step = self.get_several_member("Fem::ConstraintRigidBody")
        self.cons_force = self.get_several_member("Fem::ConstraintForce")
        self.cons_heatflux = self.get_several_member("Fem::ConstraintHeatflux")
        self.cons_initialtemperature = self.get_several_member("Fem::ConstraintInitialTemperature")
        self.cons_finaltemperature = self.get_several_member("Fem::ConstraintInitialTemperature")
        self.cons_planerotation = self.get_several_member("Fem::ConstraintPlaneRotation")
        self.cons_pressure = self.get_several_member("Fem::ConstraintPressure")
        self.cons_sectionprint = self.get_several_member("Fem::ConstraintSectionPrint")
        self.cons_selfweight = self.get_several_member("Fem::ConstraintSelfWeight")
        self.cons_temperature = self.get_several_member("Fem::ConstraintTemperature")
        self.cons_tie = self.get_several_member("Fem::ConstraintTie")
        self.cons_transform = self.get_several_member("Fem::ConstraintTransform")
        self.cons_electrostatic = self.get_several_member("Fem::ConstraintElectrostaticPotential")
        self.cons_electricchargedensity = self.get_several_member(
            "Fem::ConstraintElectricChargeDensity"
        )

    def get_several_member(self, t):
        return get_several_member(self.analysis, t)



# ***************************************************************************
# *   Copyright (c) 2002 Juergen Riegel <juergen.riegel@web.de>             *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/

import FreeCAD
import sys
import unittest


# ---------------------------------------------------------------------------
# define the functions to test the FreeCAD base code
# ---------------------------------------------------------------------------


def tryLoadingTest(testName):
    "Loads and returns testName, or a failing TestCase if unsuccessful."

    try:
        return unittest.defaultTestLoader.loadTestsFromName(testName)

    except ImportError:

        class LoadFailed(unittest.TestCase):
            def __init__(self, testName):
                # setattr() first, because TestCase ctor checks for methodName.
                setattr(self, "failed_to_load_" + testName, self._runTest)
                super(LoadFailed, self).__init__("failed_to_load_" + testName)
                self.testName = testName

            def __name__(self):
                return "Loading " + self.testName

            def _runTest(self):
                self.fail("Couldn't load " + self.testName)

        return LoadFailed(testName)


def All():
    # Registered tests
    tests = FreeCAD.__unit_test__

    suite = unittest.TestSuite()

    for test in tests:
        suite.addTest(tryLoadingTest(test))

    return suite


def PrintAll():
    # Registered tests
    tests = FreeCAD.__unit_test__

    suite = unittest.TestSuite()

    FreeCAD.Console.PrintMessage("\nRegistered test units:\n\n")
    for test in tests:
        FreeCAD.Console.PrintMessage(("%s\n" % test))
    FreeCAD.Console.PrintMessage("\nPlease choose one or use 0 for all\n")

    return suite


def TestText(s):
    s = unittest.defaultTestLoader.loadTestsFromName(s)
    r = unittest.TextTestRunner(stream=sys.stdout, verbosity=2)
    retval = r.run(s)
    # Flushing to make sure the stream is written to the console
    # before the wrapping process stops executing. Without this line
    # executing the tests from command line did not show stats
    # and proper traceback in some cases.
    sys.stdout.flush()
    return retval


def Test(s):
    TestText(s)


def testAll():
    r = unittest.TextTestRunner(stream=sys.stdout, verbosity=2)
    return r.run(All())


def testUnit():
    TestText(unittest.TestLoader().loadTestsFromName("UnitTests"))


def testDocument():
    suite = unittest.TestSuite()
    suite.addTest(unittest.defaultTestLoader.loadTestsFromName("Document"))
    TestText(suite)



# ***************************************************************************
# *   Copyright (c) 2011 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import unittest

import FreeCAD
from FreeCAD import Base
import TestSketcherApp


class TestPad(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("PartDesignTestPad")

    def testBoxCase(self):
        self.PadSketch = self.Doc.addObject("Sketcher::SketchObject", "SketchPad")
        TestSketcherApp.CreateSlotPlateSet(self.PadSketch)
        self.Doc.recompute()
        self.Pad = self.Doc.addObject("PartDesign::Pad", "Pad")
        self.Pad.Profile = self.PadSketch
        self.Doc.recompute()
        self.assertEqual(len(self.Pad.Shape.Faces), 6)

    def testSketchOnBasePlane(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.PadSketch = self.Doc.addObject("Sketcher::SketchObject", "SketchPad")
        self.PadSketch.AttachmentSupport = (self.Doc.XY_Plane, [""])
        self.PadSketch.MapMode = "FlatFace"
        self.Body.addObject(self.PadSketch)
        TestSketcherApp.CreateSlotPlateSet(self.PadSketch)
        self.Doc.recompute()
        self.Pad = self.Doc.addObject("PartDesign::Pad", "Pad")
        self.Pad.Profile = self.PadSketch
        self.Body.addObject(self.Pad)
        self.Doc.recompute()
        self.assertEqual(len(self.Pad.Shape.Faces), 6)

    def testSketchOnDatumPlane(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.DatumPlane = self.Doc.addObject("PartDesign::Plane", "DatumPlane")
        self.DatumPlane.AttachmentSupport = (self.Doc.XY_Plane, [""])
        self.DatumPlane.MapMode = "FlatFace"
        self.Body.addObject(self.DatumPlane)
        self.PadSketch = self.Doc.addObject("Sketcher::SketchObject", "SketchPad")
        self.PadSketch.AttachmentSupport = (self.DatumPlane, [""])
        self.PadSketch.MapMode = "FlatFace"
        self.Body.addObject(self.PadSketch)
        TestSketcherApp.CreateSlotPlateSet(self.PadSketch)
        self.Doc.recompute()
        self.Pad = self.Doc.addObject("PartDesign::Pad", "Pad")
        self.Pad.Profile = self.PadSketch
        self.Body.addObject(self.Pad)
        self.Doc.recompute()
        self.assertEqual(len(self.Pad.Shape.Faces), 6)

    def testPadToFirstCase(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        # Make first offset cube Pad
        self.PadSketch = self.Doc.addObject("Sketcher::SketchObject", "SketchPad")
        self.Body.addObject(self.PadSketch)
        TestSketcherApp.CreateRectangleSketch(self.PadSketch, (0, 1), (1, 1))
        self.Doc.recompute()
        self.Pad = self.Doc.addObject("PartDesign::Pad", "Pad")
        self.Body.addObject(self.Pad)
        self.Pad.Profile = self.PadSketch
        self.Pad.Length = 1
        self.Doc.recompute()
        # Make second pad on different plane and pad to first
        self.PadSketch1 = self.Doc.addObject("Sketcher::SketchObject", "SketchPad1")
        self.Body.addObject(self.PadSketch1)
        self.PadSketch1.MapMode = "FlatFace"
        self.PadSketch1.AttachmentSupport = (self.Doc.XZ_Plane, [""])
        self.Doc.recompute()
        TestSketcherApp.CreateRectangleSketch(self.PadSketch1, (0, 0), (1, 1))
        self.Doc.recompute()
        self.Pad1 = self.Doc.addObject("PartDesign::Pad", "Pad1")
        self.Body.addObject(self.Pad1)
        self.Pad1.Profile = self.PadSketch1
        self.Pad1.Type = 2
        self.Pad1.Reversed = 1
        self.Doc.recompute()
        self.assertAlmostEqual(self.Pad1.Shape.Volume, 2.0)

    def testPadtoLastCase(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        # Make first offset cube Pad
        self.PadSketch = self.Doc.addObject("Sketcher::SketchObject", "SketchPad")
        self.Body.addObject(self.PadSketch)
        TestSketcherApp.CreateRectangleSketch(self.PadSketch, (0.5, 1), (0.5, 2))
        self.Doc.recompute()
        self.Pad = self.Doc.addObject("PartDesign::Pad", "Pad")
        self.Body.addObject(self.Pad)
        self.Pad.Profile = self.PadSketch
        self.Pad.Length = 1
        self.Doc.recompute()
        # Make second pad on different plane and pad to first
        self.PadSketch1 = self.Doc.addObject("Sketcher::SketchObject", "SketchPad1")
        self.Body.addObject(self.PadSketch1)
        self.PadSketch1.MapMode = "FlatFace"
        self.PadSketch1.AttachmentSupport = (self.Doc.XZ_Plane, [""])
        self.Doc.recompute()
        TestSketcherApp.CreateRectangleSketch(self.PadSketch1, (0, 0), (1, 1))
        self.Doc.recompute()
        self.Pad1 = self.Doc.addObject("PartDesign::Pad", "Pad1")
        self.Body.addObject(self.Pad1)
        self.Pad1.Profile = self.PadSketch1
        self.Pad1.Type = 1
        self.Pad1.Reversed = 1
        self.Doc.recompute()
        self.assertAlmostEqual(self.Pad1.Shape.Volume, 3.0)

    def testPadToFaceCase(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        # Make first offset cube Pad
        self.PadSketch = self.Doc.addObject("Sketcher::SketchObject", "SketchPad")
        self.Body.addObject(self.PadSketch)
        TestSketcherApp.CreateRectangleSketch(self.PadSketch, (0, 1), (1, 1))
        self.Doc.recompute()
        self.Pad = self.Doc.addObject("PartDesign::Pad", "Pad")
        self.Body.addObject(self.Pad)
        self.Pad.Profile = self.PadSketch
        self.Pad.Length = 1
        self.Doc.recompute()
        # Make second pad on different plane and pad to face on first
        self.PadSketch1 = self.Doc.addObject("Sketcher::SketchObject", "SketchPad1")
        self.Body.addObject(self.PadSketch1)
        self.PadSketch1.MapMode = "FlatFace"
        self.PadSketch1.AttachmentSupport = (self.Doc.XZ_Plane, [""])
        self.Doc.recompute()
        TestSketcherApp.CreateRectangleSketch(self.PadSketch1, (0, 0), (1, 1))
        self.Doc.recompute()
        self.Pad1 = self.Doc.addObject("PartDesign::Pad", "Pad1")
        self.Body.addObject(self.Pad1)
        self.Pad1.Profile = self.PadSketch1
        self.Pad1.Type = 3
        self.Pad1.UpToFace = (self.Pad, ["Face3"])
        self.Pad1.Reversed = 1
        self.Doc.recompute()
        self.assertAlmostEqual(self.Pad1.Shape.Volume, 2.0)

    def testPadTwoDimensionsCase(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        # Make first offset cube Pad
        self.PadSketch = self.Doc.addObject("Sketcher::SketchObject", "SketchPad")
        self.Body.addObject(self.PadSketch)
        TestSketcherApp.CreateRectangleSketch(self.PadSketch, (0, 1), (1, 1))
        self.Doc.recompute()
        self.Pad = self.Doc.addObject("PartDesign::Pad", "Pad")
        self.Body.addObject(self.Pad)
        self.Pad.Profile = self.PadSketch
        self.Pad.Length = 1
        self.Doc.recompute()
        # Make second pad on different plane and pad to face on first
        self.PadSketch1 = self.Doc.addObject("Sketcher::SketchObject", "SketchPad1")
        self.Body.addObject(self.PadSketch1)
        self.PadSketch1.MapMode = "FlatFace"
        self.PadSketch1.AttachmentSupport = (self.Doc.XZ_Plane, [""])
        self.Doc.recompute()
        TestSketcherApp.CreateRectangleSketch(self.PadSketch1, (0, 0), (1, 1))
        self.Doc.recompute()
        self.Pad1 = self.Doc.addObject("PartDesign::Pad", "Pad1")
        self.Body.addObject(self.Pad1)
        self.Pad1.Profile = self.PadSketch1
        self.Pad1.SideType = 1
        self.Pad1.Length = 1.0
        self.Pad1.Length2 = 2.0
        self.Pad1.Reversed = 1
        self.Doc.recompute()
        self.assertAlmostEqual(self.Pad1.Shape.Volume, 4.0)

    def testPadToConcaveCase(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        # Make a half revolution
        self.RevolutionSketch = self.Doc.addObject("Sketcher::SketchObject", "SketchPad")
        self.Body.addObject(self.RevolutionSketch)
        TestSketcherApp.CreateRectangleSketch(self.RevolutionSketch, (9, 0), (10, 5))
        self.Doc.recompute()
        self.Revolution = self.Doc.addObject("PartDesign::Revolution", "Revolution")
        self.Body.addObject(self.Revolution)
        self.Revolution.Profile = self.RevolutionSketch
        self.Revolution.ReferenceAxis = (self.RevolutionSketch, ["V_Axis"])
        self.Revolution.Angle = 180
        self.Doc.recompute()
        # Make a sketch and pad to first
        self.PadSketch = self.Doc.addObject("Sketcher::SketchObject", "SketchPad")
        self.Body.addObject(self.PadSketch)
        self.Doc.recompute()
        TestSketcherApp.CreateRectangleSketch(self.PadSketch, (0, 0), (1, 1))
        self.Doc.recompute()
        self.Pad = self.Doc.addObject("PartDesign::Pad", "Pad")
        self.Body.addObject(self.Pad)
        self.Pad.Profile = self.PadSketch
        self.Pad.Type = 2
        self.Pad.Reversed = True
        self.Doc.recompute()
        self.assertAlmostEqual(self.Pad.Shape.Volume, 2208.0963, places=4)

    def testPadToShapeCase(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        # Make first offset cube Pad
        self.PadSketch = self.Doc.addObject("Sketcher::SketchObject", "SketchPad")
        self.Body.addObject(self.PadSketch)
        TestSketcherApp.CreateRectangleSketch(self.PadSketch, (0, 1), (1, 1))
        self.Doc.recompute()
        self.Pad = self.Doc.addObject("PartDesign::Pad", "Pad")
        self.Body.addObject(self.Pad)
        self.Pad.Profile = self.PadSketch
        self.Pad.Length = 1
        self.Doc.recompute()
        # Make second pad on different plane and pad to first
        self.PadSketch1 = self.Doc.addObject("Sketcher::SketchObject", "SketchPad1")
        self.Body.addObject(self.PadSketch1)
        self.PadSketch1.MapMode = "FlatFace"
        self.PadSketch1.AttachmentSupport = (self.Doc.XZ_Plane, [""])
        self.PadSketch1.AttachmentOffset.Rotation.Axis = Base.Vector(0, 1, 0)
        self.PadSketch1.AttachmentOffset.Rotation.Angle = 0.436332  # 25°
        self.PadSketch1.AttachmentOffset.Base.z = 1
        self.Doc.recompute()
        TestSketcherApp.CreateRectangleSketch(self.PadSketch1, (1, 0), (1, 1))
        self.Doc.recompute()
        self.Pad1 = self.Doc.addObject("PartDesign::Pad", "Pad1")
        self.Body.addObject(self.Pad1)
        self.Pad1.Profile = self.PadSketch1
        self.Pad1.Type = 5
        self.Doc.recompute()
        self.assertAlmostEqual(self.Pad1.Shape.Volume, 2.58787, places=4)

    def testPadToPlaneCustomDir(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        # Make first offset cube Pad
        self.PadSketch = self.Doc.addObject("Sketcher::SketchObject", "SketchPad")
        self.Body.addObject(self.PadSketch)
        TestSketcherApp.CreateRectangleSketch(self.PadSketch, (0, 1), (1, 1))
        self.Doc.recompute()
        self.Pad = self.Doc.addObject("PartDesign::Pad", "Pad")
        self.Body.addObject(self.Pad)
        self.Pad.Profile = self.PadSketch
        self.Pad.Type = 3
        self.Doc.Pad.UseCustomVector = True
        self.Doc.Pad.Direction = Base.Vector(0, 1, 1)
        self.Pad.UpToFace = (self.Doc.XZ_Plane, [""])
        self.Doc.recompute()
        self.assertAlmostEqual(self.Pad.Shape.Volume, 1.5)

    def tearDown(self):
        # closing doc
        FreeCAD.closeDocument("PartDesignTestPad")
        # print ("omit closing document for debugging")



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2025 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Wrap PySide so the same import can use either PySide6 or PySide2. Also support using the
FreeCAD wrapper, if that is available."""

try:
    from PySide import QtCore, QtGui, QtNetwork, QtSvg, QtWidgets
except ImportError:
    try:
        from PySide6 import QtCore, QtGui, QtNetwork, QtSvg, QtWidgets
    except ImportError:
        try:
            from PySide2 import QtCore, QtGui, QtNetwork, QtSvg, QtWidgets
        except ImportError:
            raise ImportError(
                "No viable version of PySide was found (tried the FreeCAD PySide wrapper, PySide6 and PySide2)"
            )

# Dummy usage so the linter doesn't complain about the unused imports (since the whole point here is
# that the imports aren't used in this file, they are just wrapped here)
if hasattr(QtCore, "silence_the_linter"):
    pass
if hasattr(QtGui, "silence_the_linter"):
    pass
if hasattr(QtNetwork, "silence_the_linter"):
    pass
if hasattr(QtSvg, "silence_the_linter"):
    pass
if hasattr(QtWidgets, "silence_the_linter"):
    pass



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import Path
from Path.Tool.toolbit.ui import ToolBitEditorPanel


class TaskPanel:
    """TaskPanel for the SetupSheet - if it is being edited directly."""

    def __init__(self, vobj, deleteOnReject):
        Path.Log.track(vobj.Object.Label)
        self.vobj = vobj
        self.obj = vobj.Object
        self.editor = ToolBitEditorPanel(self.obj, self.editor.form)
        self.deleteOnReject = deleteOnReject
        FreeCAD.ActiveDocument.openTransaction("Edit ToolBit")

    def reject(self):
        FreeCAD.ActiveDocument.abortTransaction()
        self.editor.reject()
        FreeCADGui.Control.closeDialog()
        if self.deleteOnReject:
            FreeCAD.ActiveDocument.openTransaction("Uncreate ToolBit")
            self.editor.reject()
            FreeCAD.ActiveDocument.removeObject(self.obj.Name)
            FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()

    def accept(self):
        self.editor.accept()

        FreeCAD.ActiveDocument.commitTransaction()
        FreeCADGui.ActiveDocument.resetEdit()
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()

    def updateUI(self):
        Path.Log.track()
        self.editor.updateUI()

    def updateModel(self):
        self.editor.updateTool()
        FreeCAD.ActiveDocument.recompute()

    def setupUi(self):
        self.editor.setupUI()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides various functions to work with arcs."""
## @package arcs
# \ingroup draftgeoutils
# \brief Provides various functions to work with arcs.

import math
import lazy_loader.lazy_loader as lz

import FreeCAD as App
import DraftVecUtils

from draftgeoutils.general import geomType
from draftgeoutils.edges import findMidpoint

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")

## \addtogroup draftgeoutils
# @{


def isClockwise(edge, ref=None):
    """Return True if a circle-based edge has a clockwise direction.

    Parameters
    ----------
    edge :
        The edge to be analyzed.

    ref : Vector, optional
        The normal around which the direction of the edge is to be determined.
        Defaults to the Z normal vector.

    Returns
    -------
    bool
        Returns True if the edge is clockwise oriented around the ref Vector
        or not.
    """
    if not geomType(edge) == "Circle":
        return True

    v1 = edge.Curve.tangent(edge.ParameterRange[0])[0]
    if DraftVecUtils.isNull(v1):
        return True

    # we take an arbitrary other point on the edge that has little chances
    # to be aligned with the first one
    v2 = edge.Curve.tangent(edge.ParameterRange[0] + 0.01)[0]
    n = edge.Curve.Axis
    # if that axis points "the wrong way" from the reference, we invert it
    if not ref:
        ref = App.Vector(0, 0, 1)
    if n.getAngle(ref) > math.pi / 2:
        n = n.negative()

    if DraftVecUtils.angle(v1, v2, n) < 0:
        return False

    if n.z < 0:
        return False

    return True


def isWideAngle(edge):
    """Return True if the given edge is an arc with angle > 180 degrees."""
    if geomType(edge) != "Circle":
        return False

    r = edge.Curve.Radius
    total = 2 * r * math.pi

    if edge.Length > total / 2:
        return True

    return False


def arcFrom2Pts(firstPt, lastPt, center, axis=None):
    """Build an arc with center and 2 points, can be oriented with axis."""
    radius1 = firstPt.sub(center).Length
    radius2 = lastPt.sub(center).Length

    # (PREC = 4 = same as Part Module),  Is it possible?
    if round(radius1 - radius2, 4) != 0:
        return None

    thirdPt = App.Vector(firstPt.sub(center).add(lastPt).sub(center))
    thirdPt.normalize()
    thirdPt.scale(radius1, radius1, radius1)
    thirdPt = thirdPt.add(center)
    newArc = Part.Edge(Part.Arc(firstPt, thirdPt, lastPt))

    if axis and newArc.Curve.Axis.dot(axis) < 0:
        thirdPt = thirdPt.sub(center)
        thirdPt.scale(-1, -1, -1)
        thirdPt = thirdPt.add(center)
        newArc = Part.Edge(Part.Arc(firstPt, thirdPt, lastPt))

    return newArc


def arcFromSpline(edge):
    """Turn given edge into a circular arc from three points.

    Takes its first point, midpoint and endpoint. It works best with bspline
    segments such as those from imported svg files. Use this only
    if you are sure your edge is really an arc.

    It returns None if there is a problem, including passing straight edges.
    """
    if geomType(edge) == "Line":
        print("This edge is straight, cannot build an arc on it")
        return None

    if len(edge.Vertexes) > 1:
        # 2-point arc
        p1 = edge.Vertexes[0].Point
        p2 = edge.Vertexes[-1].Point
        ml = edge.Length / 2
        p3 = edge.valueAt(ml)
        try:
            return Part.Arc(p1, p3, p2).toShape()
        except Part.OCCError:
            print("Couldn't make an arc out of this edge")
            return None
    else:
        # circle
        p1 = edge.Vertexes[0].Point
        p2 = findMidpoint(edge)
        ray = p2.sub(p1)
        ray.scale(0.5, 0.5, 0.5)
        center = p1.add(ray)
        radius = ray.Length
        try:
            return Part.makeCircle(radius, center)
        except Part.OCCError:
            print("couldn't make a circle out of this edge")
            return None


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# /***************************************************************************
# *   Copyright (c) 2016 Victor Titov (DeepSOIC) <vv.titov@gmail.com>       *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************/

__title__ = "BOPTools package"
__url__ = "https://www.freecad.org"
__doc__ = """BOPTools Package (part of FreeCAD). Routines that power Connect, Embed, Cutout,
BooleanFragments, Slice and XOR features of Part Workbench. Useful for other custom
BOP-like operations"""

## @package BOPTools
#  \ingroup PART

__all__ = [
    "BOPFeatures",
    "GeneralFuseResult",
    "JoinAPI",
    "JoinFeatures",
    "ShapeMerge",
    "Utils",
    "SplitAPI",
    "SplitFeatures",
]


def importAll():
    "importAll(): imports all modules of BOPTools package"
    from . import BOPFeatures
    from . import GeneralFuseResult
    from . import JoinAPI
    from . import JoinFeatures
    from . import ShapeMerge
    from . import Utils
    from . import SplitAPI
    from . import SplitFeatures
    from . import ToleranceFeatures


def reloadAll():
    "reloadAll(): reloads all modules of BOPTools package. Useful for debugging."
    for modstr in __all__:
        reload(globals()[modstr])
    import FreeCAD

    if FreeCAD.GuiUp:
        addCommands()


def addCommands():
    "addCommands(): add all GUI commands of BOPTools package to FreeCAD command manager."
    JoinFeatures.addCommands()
    SplitFeatures.addCommands()
    ToleranceFeatures.addCommands()
